#!/usr/bin/env python
# Copyright (c) 2023-2024, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.

"""Tests for `io.backends.furuno` module.

Ported from wradlib.
"""

import datetime
import gzip
import os
import shutil
import struct
import tempfile

import numpy as np
import pytest
from xarray import DataTree, Variable
from xarray.backends import CachingFileManager
from xarray.core.indexing import OuterIndexer
from xarray.core.utils import Frozen

from xradar.io.backends import furuno, open_furuno_datatree
from xradar.io.backends.common import _get_fmt_string
from xradar.io.backends.furuno import (
    YMDS_TIME,
    FurunoBackendEntrypoint,
    FurunoFile,
    FurunoStore,
    decode_time,
)
from xradar.util import _get_data_file


def test_open_scn(furuno_scn_file, file_or_filelike):
    with _get_data_file(furuno_scn_file, file_or_filelike) as furunofile:
        data = FurunoFile(furunofile, loaddata=False, obsmode=1)
    assert isinstance(data, furuno.FurunoFile)
    assert isinstance(data.fh, (np.memmap, np.ndarray))
    with _get_data_file(furuno_scn_file, file_or_filelike) as furunofile:
        data = furuno.FurunoFile(furunofile, loaddata=True, obsmode=1)
    assert len(data.data) == 11
    assert data.filename == furunofile
    assert data.version == 3
    assert data.a1gate == 796
    assert data.angle_resolution == 0.26
    assert data.first_dimension == "azimuth"
    assert data.fixed_angle == 7.8
    assert data.site_coords == (15.44729, 47.07734000000001, 407.9)
    assert data.header["scan_start_time"] == datetime.datetime(2021, 7, 30, 16, 0)
    assert list(data.data.keys()) == [
        "RATE",
        "DBZH",
        "VRADH",
        "ZDR",
        "KDP",
        "PHIDP",
        "RHOHV",
        "WRADH",
        "QUAL",
        "azimuth",
        "elevation",
    ]


def test_open_scn_filelike(furuno_scn_file):
    with pytest.raises(
        ValueError, match="Furuno `observation mode` can't be extracted"
    ):
        with _get_data_file(furuno_scn_file, "filelike") as furunofile:
            data = furuno.FurunoFile(furunofile, loaddata=False)
            print(data.first_dimension)


def test_open_scnx(furuno_scnx_file, file_or_filelike):
    with _get_data_file(furuno_scnx_file, file_or_filelike) as furunofile:
        data = furuno.FurunoFile(furunofile, loaddata=False)
    assert isinstance(data, furuno.FurunoFile)
    assert isinstance(data.fh, (np.memmap, np.ndarray))
    with _get_data_file(furuno_scnx_file, file_or_filelike) as furunofile:
        data = furuno.FurunoFile(furunofile, loaddata=True)
    assert data.filename == furunofile
    assert data.version == 10
    assert data.a1gate == 292
    assert data.angle_resolution == 0.5
    assert data.first_dimension == "azimuth"
    assert data.fixed_angle == 0.5
    assert data.site_coords == (13.243970000000001, 53.55478, 38.0)
    assert data.header["scan_start_time"] == datetime.datetime(2022, 3, 24, 0, 0, 1)
    assert len(data.data) == 11
    assert list(data.data.keys()) == [
        "RATE",
        "DBZH",
        "VRADH",
        "ZDR",
        "KDP",
        "PHIDP",
        "RHOHV",
        "WRADH",
        "QUAL",
        "azimuth",
        "elevation",
    ]


def test_decode_time_valid():
    # Define valid data for YMDS_TIME
    data = {"year": 2023, "month": 8, "day": 15, "hour": 14, "minute": 30, "second": 45}
    # Create binary buffer using struct.pack
    fmt = _get_fmt_string(YMDS_TIME)
    buffer = struct.pack(
        fmt,
        data["year"],
        data["month"],
        data["day"],
        data["hour"],
        data["minute"],
        data["second"],
        b"\x00",
    )

    # Call decode_time with binary data
    result = decode_time(buffer, YMDS_TIME)
    expected = datetime.datetime(2023, 8, 15, 14, 30, 45)
    assert result == expected, "Expected valid datetime object"


def test_decode_time_invalid_date():
    # Define invalid data (month 13)
    data = {
        "year": 2023,
        "month": 13,
        "day": 15,
        "hour": 14,
        "minute": 30,
        "second": 45,
    }
    # Create binary buffer for invalid date
    fmt = _get_fmt_string(YMDS_TIME)
    buffer = struct.pack(
        fmt,
        data["year"],
        data["month"],
        data["day"],
        data["hour"],
        data["minute"],
        data["second"],
        b"\x00",
    )

    # Call decode_time with invalid data
    result = decode_time(buffer, YMDS_TIME)
    assert result is None, "Expected None for an invalid date"


def prepare_temp_file_with_extension(furuno_scn_file, extension):
    # Create a temporary file with the specified extension
    fnameo = os.path.join(
        tempfile.gettempdir(), f"{os.path.basename(furuno_scn_file)[:-7]}{extension}"
    )
    with gzip.open(furuno_scn_file) as fin:
        with open(fnameo, "wb") as fout:
            shutil.copyfileobj(fin, fout)
    return fnameo


def test_furuno_file_with_binary_file(furuno_scn_file, file_or_filelike):
    fnameo = os.path.join(
        tempfile.gettempdir(), f"{os.path.basename(furuno_scn_file)[:-3]}"
    )
    with gzip.open(furuno_scn_file) as fin:
        with open(fnameo, "wb") as fout:
            shutil.copyfileobj(fin, fout)

    with furuno.FurunoFile(fnameo, loaddata=False) as data:
        # Check that _fp and _fh were set properly
        assert data._fp is not None, "_fp should be an open file object"
        assert isinstance(data._fh, np.memmap), "_fh should be a numpy memmap object"
        assert data.loaddata is False, "Expected loaddata to be False"
        assert data.rawdata is False, "Expected rawdata to be False"
        assert data.debug is False, "Expected debug to be False"
        assert data.first_dimension == "azimuth"
        assert data.filepos == 0
        data.filepos = 10
        assert data.filepos == 10
    assert data._fp.closed, "_fp should be closed after exiting the context"


def test_furuno_file_with_sppi_file(furuno_scn_file):
    fnameo = prepare_temp_file_with_extension(furuno_scn_file, ".sppi")
    with furuno.FurunoFile(fnameo, loaddata=False) as data:
        assert data.first_dimension == "azimuth", "Expected 'azimuth' for .sppi files"
    assert data._fp.closed, "_fp should be closed after exiting the context"


def test_furuno_file_with_rhi_file(furuno_scn_file):
    fnameo = prepare_temp_file_with_extension(furuno_scn_file, ".rhi")
    with furuno.FurunoFile(fnameo, loaddata=False) as data:
        assert (
            data.first_dimension == "elevation"
        ), "Expected 'elevation' for .rhi files"
    assert data._fp.closed, "_fp should be closed after exiting the context"


def test_furuno_file_unknown_obsmode(furuno_scn_file):
    # Prepare a file without a recognized extension to leave obs_mode unset
    fnameo = prepare_temp_file_with_extension(furuno_scn_file, ".txt")
    # Initialize FurunoFile without a recognized extension and no obsmode
    with pytest.raises(TypeError, match="Unknown Furuno Observation Mode: None"):
        with furuno.FurunoFile(fnameo, loaddata=False) as data:
            _ = data.first_dimension  # Access first_dimension to trigger the TypeError


def test_furuno_array_wrapper_initialization():
    # Initialize with a sample array
    data = np.arange(20).reshape(4, 5)
    wrapper = furuno.FurunoArrayWrapper(data)

    # Check that data, shape, and dtype are correctly set
    assert np.array_equal(wrapper.data, data), "Data should match the input array"
    assert wrapper.shape == (4, 5), "Shape should match the input data's shape"
    assert wrapper.dtype == np.dtype("uint16"), "Expected dtype to be uint16"


def test_furuno_array_wrapper_getitem():
    # Initialize with a sample array
    data = np.arange(20).reshape(4, 5)
    wrapper = furuno.FurunoArrayWrapper(data)

    # Use __getitem__ to access specific elements and slices
    assert (
        wrapper[OuterIndexer((1, 2))] == data[1, 2]
    ), "Single element access should match"
    assert np.array_equal(
        wrapper[OuterIndexer((1, slice(None)))], data[1, :]
    ), "Row slice should match"
    assert np.array_equal(
        wrapper[OuterIndexer((slice(None), 2))], data[:, 2]
    ), "Column slice should match"


def test_furuno_array_wrapper_raw_indexing_method():
    # Initialize with a sample array
    data = np.arange(20).reshape(4, 5)
    wrapper = furuno.FurunoArrayWrapper(data)

    # Access elements using _raw_indexing_method directly
    assert (
        wrapper._raw_indexing_method((1, 2)) == data[1, 2]
    ), "Direct element access should match"
    assert np.array_equal(
        wrapper._raw_indexing_method((1, slice(None))), data[1, :]
    ), "Direct row slice should match"
    assert np.array_equal(
        wrapper._raw_indexing_method((slice(None), 2)), data[:, 2]
    ), "Direct column slice should match"


def test_furuno_store_initialization(furuno_scn_file):
    # Initialize FurunoStore with a CachingFileManager
    manager = CachingFileManager(furuno.FurunoFile, furuno_scn_file)
    store = furuno.FurunoStore(manager)
    assert isinstance(store, furuno.FurunoStore), "Expected FurunoStore instance"
    assert store._manager == manager, "Expected manager to be set"
    assert store._group is None, "Expected group to be None by default"
    assert isinstance(store.root, furuno.FurunoFile)
    assert store.filename == furuno_scn_file, "Filename should match input file"
    assert isinstance(
        store.ds, furuno.FurunoFile
    ), "Expected ds property to return a FurunoFile instance"
    # Assuming 'DBZH' is a variable in the dataset
    var_data = np.random.rand(10, 10)  # Dummy data for testing
    dim = store.root.first_dimension
    variable = store.open_store_variable("DBZH", var_data)

    assert isinstance(variable, Variable), "Expected a Variable instance"
    assert variable.dims == (dim, "range"), "Expected dimensions to match"
    assert "scale_factor" in variable.attrs, "Expected scale_factor in attributes"


def test_furuno_store_open(furuno_scn_file):
    # Use the `open` method to initialize a FurunoStore instance
    store = FurunoStore.open(furuno_scn_file, mode="r", group="test_group")
    # Check that the returned object is an instance of FurunoStore
    assert isinstance(store, FurunoStore), "Expected a FurunoStore instance"
    # Verify that manager is an instance of CachingFileManager
    assert isinstance(
        store._manager, CachingFileManager
    ), "Expected a CachingFileManager instance"
    # Use acquire_context to confirm it manages FurunoFile instances
    with store._manager.acquire_context() as file_obj:
        assert isinstance(
            file_obj, FurunoFile
        ), "Expected CachingFileManager to manage FurunoFile instances"
    # Ensure the filename and group are set correctly
    assert store._group == "test_group", "Expected group to be 'test_group'"
    assert store.filename == furuno_scn_file, "Expected filename to match input file"


def test_furuno_store_open_store_coordinates(furuno_scn_file):
    manager = CachingFileManager(FurunoFile, furuno_scn_file)
    store = FurunoStore(manager)
    coords = store.open_store_coordinates()

    assert isinstance(coords, dict), "Expected coordinates to be a dictionary"
    assert "latitude" in coords, "Expected 'latitude' in coordinates"
    assert "longitude" in coords, "Expected 'longitude' in coordinates"
    assert "altitude" in coords, "Expected 'altitude' in coordinates"
    assert isinstance(coords["latitude"], Variable), "Expected Variable for 'latitude'"


def test_furuno_store_get_variables(furuno_scn_file):
    manager = CachingFileManager(FurunoFile, furuno_scn_file)
    store = FurunoStore(manager)
    variables = store.get_variables()

    assert isinstance(
        variables, Frozen
    ), "Expected get_variables to return a dictionary"
    assert "latitude" in variables, "Expected 'latitude' in variables"
    assert "longitude" in variables, "Expected 'longitude' in variables"


def test_furuno_store_get_attrs(furuno_scn_file):
    manager = CachingFileManager(FurunoFile, furuno_scn_file)
    store = FurunoStore(manager)
    attrs = store.get_attrs()

    assert isinstance(attrs, Frozen), "Expected attributes to be a dictionary"
    assert attrs["source"] == "Furuno", "Expected source attribute to be 'Furuno'"
    assert "version" in attrs, "Expected version attribute in attrs"


def test_furuno_store_get_calibration_parameters(furuno_scn_file):
    manager = CachingFileManager(FurunoFile, furuno_scn_file)
    store = FurunoStore(manager)
    calibration_params = store.get_calibration_parameters()
    assert isinstance(
        calibration_params, Frozen
    ), "Expected calibration parameters to be a dictionary"


def test_open_store_coordinates_version_else(furuno_scn_file):
    # Set up a mocked version of FurunoFile with header values that activate the else statement
    manager = CachingFileManager(FurunoFile, furuno_scn_file)
    store = FurunoStore(manager)

    # Mock attributes including format_version
    store.ds._header = {
        "format_version": 2,  # Ensure format_version is included
        "resolution_range_direction": 100.0,
        "number_range_direction_data": 100,
        "scan_start_time": datetime.datetime(2021, 7, 30, 16, 0, 0),
        "number_sweep_direction_data": 360,
        "antenna_rotation_speed": 10,
        "observation_mode": 1,
        "longitude": 1544729,
        "latitude": 4707734,
        "altitude": 4079,
    }

    # Mock ds.version to a value not in [3, 103]
    store.ds._version = (
        2  # Any value not 3 or 103 to trigger the else condition for range_step
    )

    # Call the function
    coords = store.open_store_coordinates()

    # Check if the range step has been applied as expected
    range_step = store.ds.header["resolution_range_direction"]
    expected_rng = np.arange(
        range_step / 2,
        range_step * store.ds.header["number_range_direction_data"] + range_step / 2,
        range_step,
        dtype="float32",
    )
    assert np.array_equal(
        coords["range"].values, expected_rng
    ), "Range values do not match expected values"


def test_open_store_coordinates_time_else(furuno_scn_file):
    # Set up a mocked version of FurunoFile with header values that trigger time-based else
    manager = CachingFileManager(FurunoFile, furuno_scn_file)
    store = FurunoStore(manager)

    # Mock attributes with missing scan_stop_time
    store.ds._header = {
        "format_version": 2,  # Ensure format_version is included
        "resolution_range_direction": 100.0,
        "number_range_direction_data": 100,
        "scan_start_time": datetime.datetime(2021, 7, 30, 16, 0, 0),
        "number_sweep_direction_data": 360,
        "antenna_rotation_speed": 10,
        "observation_mode": 1,
        "longitude": 1544729,
        "latitude": 4707734,
        "altitude": 4079,
    }

    # Mock ds.version to a value that uses resolution_range_direction directly
    store.ds._version = 3

    # Call the function
    coords = store.open_store_coordinates()

    # Calculate expected time step if no stop_time
    expected_raytime = datetime.timedelta(
        seconds=store.ds.angle_resolution
        / (store.ds.header["antenna_rotation_speed"] * 1e-1 * 6)
    )

    # Apply the initial offset
    initial_offset = coords["time"].values[0]
    expected_raytimes = np.array(
        [
            initial_offset + (x * expected_raytime).total_seconds()
            for x in range(store.ds.header["number_sweep_direction_data"])
        ]
    )

    # Check if the calculated ray times match the returned time coordinates with tolerance
    np.testing.assert_almost_equal(
        coords["time"].values, expected_raytimes, decimal=6
    ), "Time values do not match expected values"


def test_open_store_coordinates_time_else_with_stop_time(furuno_scn_file):
    # Set up a mocked version of FurunoFile with header values that trigger the else condition with stop_time
    manager = CachingFileManager(FurunoFile, furuno_scn_file)
    store = FurunoStore(manager)

    # Mock attributes including scan_stop_time to activate the else condition
    store.ds._header = {
        "format_version": 2,
        "resolution_range_direction": 100.0,
        "number_range_direction_data": 100,
        "scan_start_time": datetime.datetime(2021, 7, 30, 16, 0, 0),
        "scan_stop_time": datetime.datetime(2021, 7, 30, 16, 10, 0),  # 10 minutes later
        "number_sweep_direction_data": 360,
        "antenna_rotation_speed": 10,
        "observation_mode": 1,
        "longitude": 1544729,
        "latitude": 4707734,
        "altitude": 4079,
    }

    # Mock ds.version to use the else condition for range_step
    store.ds._version = 3

    # Call the function
    coords = store.open_store_coordinates()

    # Calculate expected ray times using the scan_stop_time for the else case
    time = store.ds.header["scan_start_time"]
    stop_time = store.ds.header["scan_stop_time"]
    num_rays = store.ds.header["number_sweep_direction_data"]
    raytime = (stop_time - time) / num_rays

    # Generate expected raytimes based on raytime intervals
    expected_raytimes = np.array(
        [((x + 0.5) * raytime).total_seconds() for x in range(num_rays)]
    )

    # Check if the calculated ray times match the returned time coordinates with tolerance
    np.testing.assert_almost_equal(
        coords["time"].values, expected_raytimes, decimal=5
    ), "Time values do not match expected values"


def test_open_dataset(furuno_scn_file):
    # Initialize FurunoBackendEntrypoint
    backend_entrypoint = furuno.FurunoBackendEntrypoint()

    # Define parameters for opening the dataset
    mask_and_scale = True
    decode_times = True
    concat_characters = True
    decode_coords = True
    drop_variables = None
    use_cftime = None
    decode_timedelta = None
    group = None
    obsmode = 1

    # Call open_dataset directly
    ds = backend_entrypoint.open_dataset(
        furuno_scn_file,
        mask_and_scale=mask_and_scale,
        decode_times=decode_times,
        concat_characters=concat_characters,
        decode_coords=decode_coords,
        drop_variables=drop_variables,
        use_cftime=use_cftime,
        decode_timedelta=decode_timedelta,
        group=group,
        obsmode=obsmode,
    )

    # Assertions to verify dataset content
    assert "azimuth" in ds.coords, "Azimuth coordinate is missing."
    assert "elevation" in ds.coords, "Elevation coordinate is missing."
    assert "time" in ds.coords, "Time coordinate is missing."

    # Check encoding
    assert ds.encoding["engine"] == "furuno", "Engine encoding mismatch."

    # Check dimension order and size
    assert list(ds.dims) == [
        "azimuth",
        "range",
    ], "Dimensions do not match expected values."
    assert ds["DBZH"].shape == (1376, 602), "DBZH shape does not match expected values."

    # Check attributes for calibration parameters and source
    assert "source" in ds.attrs, "Source attribute missing."
    assert "version" in ds.attrs, "Version attribute missing."
    assert ds.attrs["source"] == "Furuno", "Source attribute mismatch."

    # Check if reindexing and coordinate assignments are handled
    assert "latitude" in ds.coords, "Latitude coordinate missing."
    assert "longitude" in ds.coords, "Longitude coordinate missing."
    assert "altitude" in ds.coords, "Altitude coordinate missing."


def test_open_dataset_with_reindex_and_dimensions(furuno_scn_file):
    # Initialize FurunoBackendEntrypoint
    backend_entrypoint = FurunoBackendEntrypoint()

    # Define parameters to trigger all conditions
    mask_and_scale = True
    decode_times = True
    concat_characters = True
    decode_coords = True
    drop_variables = None
    use_cftime = None
    decode_timedelta = None
    group = None
    obsmode = 1
    reindex_angle = {
        "start_angle": 0.0,
        "stop_angle": 360.0,
        "angle_res": 1.0,  # Set a valid angle resolution
        "direction": 1,
    }  # Not False to trigger the reindexing block
    first_dim = "auto"  # Set to "auto" initially to test first_dim handling

    # Call open_dataset directly
    ds = backend_entrypoint.open_dataset(
        furuno_scn_file,
        mask_and_scale=mask_and_scale,
        decode_times=decode_times,
        concat_characters=concat_characters,
        decode_coords=decode_coords,
        drop_variables=drop_variables,
        use_cftime=use_cftime,
        decode_timedelta=decode_timedelta,
        group=group,
        reindex_angle=reindex_angle,  # Pass reindex_angle to trigger reindexing
        first_dim=first_dim,  # Test first_dim="auto"
        obsmode=obsmode,
    )

    # Assertions to ensure reindexing and dimension handling was applied
    assert "azimuth" in ds.coords, "Azimuth coordinate is missing."
    assert "elevation" in ds.coords, "Elevation coordinate is missing."
    assert "time" in ds.coords, "Time coordinate is missing."

    # Check if azimuth coordinate matches the expected start, stop, and resolution from reindexing
    expected_azimuth = np.linspace(
        reindex_angle["start_angle"] + reindex_angle["angle_res"] / 2,
        reindex_angle["stop_angle"] - reindex_angle["angle_res"] / 2,
        len(ds["azimuth"]),
    )
    np.testing.assert_almost_equal(
        ds["azimuth"].values,
        expected_azimuth,
        decimal=5,
        err_msg="Azimuth coordinate does not match expected reindexed values",
    )

    # Test first_dim behavior
    assert "azimuth" in ds.DBZH.dims

    # Repeat with first_dim set to "time" to check other handling
    ds_time_dim = backend_entrypoint.open_dataset(
        furuno_scn_file,
        mask_and_scale=mask_and_scale,
        decode_times=decode_times,
        concat_characters=concat_characters,
        decode_coords=decode_coords,
        drop_variables=drop_variables,
        use_cftime=use_cftime,
        decode_timedelta=decode_timedelta,
        group=group,
        reindex_angle=reindex_angle,
        first_dim="time",  # Set to "time" to test alternative branch
        obsmode=obsmode,
    )

    # Ensure that time is now the primary dimension, if expected
    assert (
        "time" in ds_time_dim.DBZH.dims
    ), "Time dimension was not set as the primary dimension when expected."


def test_open_furuno_datatree(furuno_scn_file):
    # Define parameters to trigger all conditions
    kwargs = {
        "first_dim": "auto",
        "reindex_angle": {
            "start_angle": 0.0,
            "stop_angle": 360.0,
            "angle_res": 1.0,
            "direction": 1,
        },
        "fix_second_angle": True,
        "site_coords": True,
    }

    # Call the function with the actual file
    dtree = open_furuno_datatree(furuno_scn_file, **kwargs)

    # Assertions
    assert isinstance(dtree, DataTree), "Expected a DataTree instance"
    assert "/" in dtree.subtree, "Root group should be present in the DataTree"
    assert (
        "/radar_parameters" in dtree.subtree
    ), "Radar parameters group should be in the DataTree"
    assert (
        "/georeferencing_correction" in dtree.subtree
    ), "Georeferencing correction group should be in the DataTree"
    assert (
        "/radar_calibration" in dtree.subtree
    ), "Radar calibration group should be in the DataTree"

    # Check if at least one sweep group is attached (e.g., "/sweep_0")
    sweep_groups = [key for key in dtree.match("sweep_*")]
    assert len(sweep_groups) == 1, "Expected at least one sweep group in the DataTree"

    # Verify a sample variable in one of the sweep groups (adjust as needed based on expected variables)
    sample_sweep = sweep_groups[0]
    assert (
        "DBZH" in dtree[sample_sweep].data_vars
    ), f"DBZH should be a data variable in {sample_sweep}"
    assert (
        "VRADH" in dtree[sample_sweep].data_vars
    ), f"VRADH should be a data variable in {sample_sweep}"

    # Validate coordinates are attached correctly
    assert (
        "latitude" in dtree[sample_sweep]
    ), "Latitude should be attached to the root dataset"
    assert (
        "longitude" in dtree[sample_sweep]
    ), "Longitude should be attached to the root dataset"
    assert (
        "altitude" in dtree[sample_sweep]
    ), "Altitude should be attached to the root dataset"
    assert len(dtree[sample_sweep].variables) == 21
    assert dtree[sample_sweep]["DBZH"].shape == (360, 602)
    assert len(dtree.attrs) == 9
    assert dtree.attrs["version"] == 3
    assert dtree.attrs["source"] == "Furuno"



#!/usr/bin/env python
# Copyright (c) 2024, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.


import tarfile

import pytest
from xarray import DataTree

from xradar.io.backends import datamet, open_datamet_datatree
from xradar.util import _get_data_file


@pytest.fixture(scope="session")
def data(datamet_file):
    with _get_data_file(datamet_file, "file") as datametfile:
        data = datamet.DataMetFile(datametfile)
        assert data.filename == datametfile
    data.get_sweep(0)
    return data


def test_file_types(data):
    assert isinstance(data, datamet.DataMetFile)
    assert isinstance(data.tarfile, tarfile.TarFile)


def test_basic_content(data):
    assert data.moments == ["UZ", "CZ", "V", "W", "ZDR", "PHIDP", "RHOHV", "KDP"]
    assert len(data.data[0]) == len(data.moments)
    assert data.first_dimension == "azimuth"
    assert data.scan_metadata["origin"] == "ILMONTE"
    assert data.scan_metadata["orig_lat"] == 41.9394
    assert data.scan_metadata["orig_lon"] == 14.6208
    assert data.scan_metadata["orig_alt"] == 710


def test_moment_metadata(data):
    mom_metadata = data.get_mom_metadata("UZ", 0)
    assert mom_metadata["Rangeoff"] == 0.0
    assert mom_metadata["Eloff"] == 16.05
    assert mom_metadata["nlines"] == 360
    assert mom_metadata["ncols"] == 493


@pytest.mark.parametrize(
    "moment, expected_value",
    [
        ("UZ", 56),
        ("CZ", 56),
        ("V", 139),
        ("W", 30),
        ("ZDR", 137),
        ("PHIDP", 16952),
        ("RHOHV", 237),
        ("KDP", 67),
    ],
)
def test_moment_data(data, moment, expected_value):
    assert data.data[0][moment][(4, 107)] == expected_value


def test_open_datamet_datatree(datamet_file):
    # Define kwargs to pass into the function
    kwargs = {
        "sweep": [0, 1],  # Test with specific sweeps
        "first_dim": "auto",
        "reindex_angle": {
            "start_angle": 0.0,
            "stop_angle": 360.0,
            "angle_res": 1.0,
            "direction": 1,  # Set a valid direction within reindex_angle
        },
        "site_coords": True,
    }

    # Call the function with an actual DataMet file
    dtree = open_datamet_datatree(datamet_file, **kwargs)

    assert isinstance(dtree, DataTree), "Expected a DataTree instance"
    assert "/" in dtree.subtree, "Root group should be present in the DataTree"
    assert (
        "/radar_parameters" in dtree.subtree
    ), "Radar parameters group should be in the DataTree"
    assert (
        "/georeferencing_correction" in dtree.subtree
    ), "Georeferencing correction group should be in the DataTree"
    assert (
        "/radar_calibration" in dtree.subtree
    ), "Radar calibration group should be in the DataTree"

    # Check if at least one sweep group is attached (e.g., "/sweep_0")
    sweep_groups = [key for key in dtree.match("sweep_*")]
    assert len(sweep_groups) == 2, "Expected four sweep groups in the DataTree"

    # Verify a sample variable in one of the sweep groups (adjust based on expected variables)
    sample_sweep = sweep_groups[0]
    assert (
        len(dtree[sample_sweep].data_vars) == 13
    ), f"Expected data variables in {sample_sweep}"
    assert dtree[sample_sweep]["DBZH"].shape == (360, 493)
    assert (
        "DBZH" in dtree[sample_sweep].data_vars
    ), f"DBZH should be a data variable in {sample_sweep}"
    assert (
        "VRADH" in dtree[sample_sweep].data_vars
    ), f"VRADH should be a data variable in {sample_sweep}"

    # Validate coordinates are attached correctly in the root dataset
    assert (
        "latitude" in dtree[sample_sweep]
    ), "Latitude should be attached to the root dataset"
    assert (
        "longitude" in dtree[sample_sweep]
    ), "Longitude should be attached to the root dataset"
    assert (
        "altitude" in dtree[sample_sweep]
    ), "Altitude should be attached to the root dataset"

    # Validate attributes
    assert len(dtree.attrs) == 10
    assert (
        dtree.attrs["instrument_name"] == "ILMONTE"
    ), "Instrument name should match expected value"
    assert dtree.attrs["source"] == "Datamet", "Source should match expected value"
    assert (
        dtree.attrs["scan_name"] == "VOLUMETRICA"
    ), "Scan name should match expected value"

    # Verify a sample variable in on



#!/usr/bin/env python
# Copyright (c) 2024, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.

"""Tests for `io.backends.gamic` module.

ported from wradlib
"""

import numpy as np
import pytest
from xarray import DataTree

from xradar.io.backends import gamic
from xradar.io.backends.gamic import open_gamic_datatree


def create_ray_header(nrays=360):
    dtype = {
        "names": [
            "azimuth_start",
            "azimuth_stop",
            "elevation_start",
            "elevation_stop",
            "timestamp",
        ],
        "formats": ["<f8", "<f8", "<f8", "<f8", "<i8"],
        "offsets": [0, 8, 16, 24, 32],
        "itemsize": 40,
    }
    ray_header = np.zeros(nrays, dtype=dtype)
    start_az = np.linspace(0, 360, nrays, endpoint=False, dtype=np.float64)
    stop_az = np.linspace(1, 361, nrays, endpoint=False, dtype=np.float64)
    ray_header["azimuth_start"] = start_az
    ray_header["azimuth_stop"] = stop_az

    start_el = np.ones_like(start_az, dtype=np.float64)
    stop_el = np.ones_like(start_az, dtype=np.float64)
    ray_header["elevation_start"] = start_el
    ray_header["elevation_stop"] = stop_el

    time = np.arange(
        1527831788042000, 1527831788042000 + nrays * 100000, 100000, dtype=np.int64
    )
    ray_header["timestamp"] = time
    return ray_header


@pytest.mark.parametrize("nrays", [180, 240, 360, 720])
def test_get_azimuth(nrays):
    ray_header = create_ray_header(nrays)
    actual = gamic._get_azimuth(ray_header)
    udiff = np.unique(np.diff(actual))
    assert len(actual) == nrays
    assert len(udiff) == 1
    assert udiff[0] == 360.0 / nrays


@pytest.mark.parametrize("nrays", [180, 240, 360, 720])
def test_get_elevation(nrays):
    ray_header = create_ray_header(nrays)
    actual = gamic._get_elevation(ray_header)
    unique = np.unique(actual)
    assert len(actual) == nrays
    assert len(unique) == 1
    assert unique[0] == 1.0


@pytest.mark.parametrize("nrays", [180, 240, 360, 720])
def test_get_timestamp(nrays):
    ray_header = create_ray_header(nrays)
    actual = gamic._get_time(ray_header)
    udiff = np.unique(np.diff(actual))
    assert len(actual) == nrays
    assert len(udiff) == 1
    assert udiff[0] == 100000


@pytest.mark.parametrize(
    "ang",
    [("elevation", "azimuth"), ("azimuth", "elevation")],
)
def test_get_fixed_dim_and_angle(ang):
    how = {ang[0]: 1.0}
    dim, angle = gamic._get_fixed_dim_and_angle(how)
    assert dim == ang[1]
    assert angle == 1.0


@pytest.mark.parametrize("range_step", [100, 150, 300, 1000])
@pytest.mark.parametrize("range_samples", [1, 2, 4])
def test_get_range(range_step, range_samples):
    where = dict(bin_count=10, range_step=range_step, range_samples=range_samples)
    rng, cent_first, bin_range = gamic._get_range(where)
    assert len(rng) == 10
    assert np.unique(np.diff(rng))[0] == range_step * range_samples
    assert cent_first == (range_step * range_samples) / 2
    assert bin_range == range_step * range_samples


def test_GamicH5NetCDFMetadata(gamic_file):
    store = gamic.GamicStore.open(gamic_file, group="sweep_0")
    with pytest.warns(DeprecationWarning):
        assert store.root.first_dim == "azimuth"


def test_open_gamic_datatree(gamic_file):
    # Define kwargs to pass into the function
    kwargs = {
        "sweep": [0, 1],  # Test with specific sweeps
        "first_dim": "auto",
        "reindex_angle": {
            "start_angle": 0.0,
            "stop_angle": 360.0,
            "angle_res": 1.0,
            "direction": 1,  # Set a valid direction within reindex_angle
        },
        "fix_second_angle": True,
        "site_coords": True,
    }

    # Call the function with an actual GAMIC HDF5 file
    dtree = open_gamic_datatree(gamic_file, **kwargs)

    # Assertions
    assert isinstance(dtree, DataTree), "Expected a DataTree instance"
    assert "/" in dtree.subtree, "Root group should be present in the DataTree"
    assert (
        "/radar_parameters" in dtree.subtree
    ), "Radar parameters group should be in the DataTree"
    assert (
        "/georeferencing_correction" in dtree.subtree
    ), "Georeferencing correction group should be in the DataTree"
    assert (
        "/radar_calibration" in dtree.subtree
    ), "Radar calibration group should be in the DataTree"

    # Check if at least one sweep group is attached (e.g., "/sweep_0")
    sweep_groups = [key for key in dtree.match("sweep_*")]
    assert len(sweep_groups) == 2, "Expected two sweep groups in the DataTree"

    # Verify a sample variable in one of the sweep groups
    sample_sweep = sweep_groups[0]
    assert (
        len(dtree[sample_sweep].data_vars) == 17
    ), f"Expected data variables in {sample_sweep}"
    assert dtree[sample_sweep]["DBZH"].shape == (360, 360)
    assert (
        "DBZH" in dtree[sample_sweep].data_vars
    ), f"DBZH should be a data variable in {sample_sweep}"
    assert (
        "VRADV" in dtree[sample_sweep].data_vars
    ), f"VRADV should be a data variable in {sample_sweep}"

    # Validate coordinates are attached correctly in the root dataset
    assert (
        "latitude" in dtree[sample_sweep]
    ), "Latitude should be attached to the root dataset"
    assert (
        "longitude" in dtree[sample_sweep]
    ), "Longitude should be attached to the root dataset"
    assert (
        "altitude" in dtree[sample_sweep]
    ), "Altitude should be attached to the root dataset"

    # Validate attributes
    assert len(dtree.attrs) == 9
    assert dtree.attrs["source"] == "gamic", "Source should match expected value"



#!/usr/bin/env python
# Copyright (c) 2022-2024, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.

"""Tests for `xradar` util package."""

import numpy as np
import pytest
import xarray as xr
from open_radar_data import DATASETS

import xradar as xd
from xradar import io, model, util
from xradar.io.backends.common import (
    _get_radar_calibration,
    _get_required_root_dataset,
    _get_subgroup,
)


@pytest.fixture(
    params=[("PPI", "azimuth", "elevation"), ("RHI", "elevation", "azimuth")]
)
def sweep(request):
    return request.param


def test_get_first_angle(sweep):
    ds = model.create_sweep_dataset(sweep=sweep[0])
    with pytest.raises(ValueError):
        util.get_first_angle(ds)
    ds = ds.swap_dims({"time": sweep[1]})
    assert util.get_first_angle(ds) == sweep[1]


def test_get_second_angle(sweep):
    ds = model.create_sweep_dataset(sweep=sweep[0])
    ds = ds.swap_dims({"time": sweep[1]})
    assert util.get_second_angle(ds) == sweep[2]


def test_remove_duplicate_rays():
    filename = DATASETS.fetch("DWD-Vol-2_99999_20180601054047_00.h5")
    ds = xr.open_dataset(filename, group="sweep_7", engine="gamic", first_dim="auto")
    ds_out = util.remove_duplicate_rays(ds)
    assert ds.sizes["azimuth"] == 361
    assert ds_out.sizes["azimuth"] == 360


def test_reindex_angle():
    filename = DATASETS.fetch("DWD-Vol-2_99999_20180601054047_00.h5")
    ds = xr.open_dataset(filename, group="sweep_7", engine="gamic", first_dim="auto")
    ds_out = util.remove_duplicate_rays(ds)
    ds_out = util.reindex_angle(
        ds_out, start_angle=0, stop_angle=360, angle_res=1.0, direction=1
    )
    assert ds_out.sizes["azimuth"] == 360
    np.testing.assert_array_equal(ds_out.azimuth.values, np.arange(0.5, 360, 1.0))


def test_extract_angle_parameters():
    filename = DATASETS.fetch("DWD-Vol-2_99999_20180601054047_00.h5")
    ds = xr.open_dataset(filename, group="sweep_7", engine="gamic", first_dim="auto")
    angle_dict = util.extract_angle_parameters(ds)
    expected_dict = {
        "a1gate_idx": np.array(243),
        "a1gate_val": np.array(243.52020263671875),
        "angle_res": np.array(1.0),
        "angles_are_unique": False,
        "ascending": np.array(True),
        "direction": 1,
        "excess_rays": np.array(True),
        "expected_angle_span": 360,
        "expected_number_rays": 360,
        "first_angle": "azimuth",
        "max_angle": np.array(359.5111083984375),
        "max_time": np.datetime64("2018-06-01T05:43:08.042000000"),
        "min_angle": np.array(0.52459716796875),
        "min_time": np.datetime64("2018-06-01T05:42:44.042000000"),
        "missing_rays": np.array(False),
        "second_angle": "elevation",
        "start_angle": 0,
        "stop_angle": 360,
        "times_are_unique": True,
        "uniform_angle_spacing": False,
    }
    assert angle_dict == expected_dict


@pytest.mark.parametrize("low", [-0.25, -0.1, 0])
@pytest.mark.parametrize("high", [0.25, 0.1, 0])
def test_extract_angle_parameters2(low, high):
    seed = 12345
    rng = np.random.default_rng(seed)
    noise = rng.uniform(low=low, high=high, size=360)
    azimuth = np.arange(0.5, 360.0, 1.0) + noise
    ds = model.create_sweep_dataset(a1gate=0, direction=1, azimuth=azimuth)
    ds = ds.swap_dims({"time": "azimuth"})
    ds = ds.sortby("azimuth")
    ds = ds.assign_coords(sweep_mode="azimuth_surveillance")

    # remove block of rays
    ds_in = xr.concat(
        [
            ds.isel(azimuth=slice(None, 100)),
            ds.isel(azimuth=slice(110, None)),
        ],
        "azimuth",
        data_vars="minimal",
    )
    angle_dict = util.extract_angle_parameters(ds_in)
    expected_dict = {
        "first_angle": "azimuth",
        "second_angle": "elevation",
        "min_angle": np.array(0.36366801123358483),
        "max_angle": np.array(359.6511639563352),
        "min_time": np.datetime64("2022-08-27T10:00:00.000000000"),
        "max_time": np.datetime64("2022-08-27T10:01:29.750000000"),
        "angles_are_unique": True,
        "times_are_unique": True,
        "a1gate_idx": np.array(0),
        "a1gate_val": np.array(0.36366801123358483),
        "uniform_angle_spacing": False,
        "ascending": np.array(True),
        "direction": 1,
        "angle_res": np.array(1.0),
        "start_angle": 0,
        "stop_angle": 360,
        "expected_angle_span": 360,
        "missing_rays": True,
        "excess_rays": False,
        "expected_number_rays": 360,
    }
    expected = {
        (0.25, -0.25): {
            "min_angle": np.array(0.36366801123358483),
            "max_angle": np.array(359.6511639563352),
            "a1gate_val": np.array(0.36366801123358483),
        },
        (0.25, -0.1): {
            "min_angle": np.array(0.4795676078635094),
            "max_angle": np.array(359.6808147694347),
            "a1gate_val": np.array(0.4795676078635094),
        },
        (0.25, 0.0): {
            "min_angle": np.array(0.5568340056167924),
            "max_angle": np.array(359.70058197816763),
            "a1gate_val": np.array(0.5568340056167924),
        },
        (0.1, -0.25): {
            "min_angle": np.array(0.32956760786350936),
            "max_angle": np.array(359.53081476943464),
            "a1gate_val": np.array(0.32956760786350936),
        },
        (0.1, -0.1): {
            "min_angle": np.array(0.4454672044934339),
            "max_angle": np.array(359.5604655825341),
            "a1gate_val": np.array(0.4454672044934339),
        },
        (0.1, 0.0): {
            "min_angle": np.array(0.5227336022467169),
            "max_angle": np.array(359.58023279126706),
            "a1gate_val": np.array(0.5227336022467169),
        },
        (0.0, -0.25): {
            "min_angle": np.array(0.3068340056167924),
            "max_angle": np.array(359.45058197816763),
            "a1gate_val": np.array(0.3068340056167924),
        },
        (0.0, -0.1): {
            "min_angle": np.array(0.42273360224671697),
            "max_angle": np.array(359.48023279126704),
            "a1gate_val": np.array(0.42273360224671697),
        },
        (0.0, 0.0): {
            "min_angle": np.array(0.5),
            "max_angle": np.array(359.5),
            "a1gate_val": np.array(0.5),
        },
    }
    for k in angle_dict:
        if k in ["min_angle", "max_angle", "a1gate_val"]:
            assert angle_dict[k] == expected[(high, low)][k]
            continue
        assert angle_dict[k] == expected_dict[k]


def test_ipol_time():
    filename = DATASETS.fetch("DWD-Vol-2_99999_20180601054047_00.h5")
    ds = xr.open_dataset(filename, group="sweep_7", engine="gamic", first_dim="auto")

    # remove block of rays (100-150)
    ds_in = xr.concat(
        [
            ds.isel(azimuth=slice(None, 100)),
            ds.isel(azimuth=slice(150, None)),
        ],
        "azimuth",
        data_vars="minimal",
    )

    # fix sweep and interpolate times
    ds_out = util.remove_duplicate_rays(ds_in)
    ds_out = util.reindex_angle(
        ds_out, start_angle=0, stop_angle=360, angle_res=1.0, direction=1
    )
    ds_out = util.ipol_time(ds_out)

    # extract times and compare
    time_in = util.remove_duplicate_rays(ds).time.astype(int)
    time_out = ds_out.time.astype(int)
    xr.testing.assert_allclose(
        time_out.drop_vars(time_out.coords), time_in.drop_vars(time_in.coords)
    )


@pytest.mark.parametrize("missing", [0, 1, 2, 10, 50])
@pytest.mark.parametrize("a1gate", [0, 1, 2, 179, 180, 181, 357, 358, 359])
@pytest.mark.parametrize("rot", ["cw", "ccw"])
def test_ipol_time2(missing, a1gate, rot):
    # prepare Dataset with one NaT in between
    if rot == "cw":
        start_ang, stop_ang = 0, 360
        direction = 1
    else:
        start_ang, stop_ang = 360, 0
        direction = -1

    ds = model.create_sweep_dataset(
        a1gate=a1gate, direction=direction, start_ang=start_ang, stop_ang=stop_ang
    )
    ds = ds.swap_dims({"time": "azimuth"})
    ds = ds.sortby("azimuth")
    ds_in = xr.concat(
        [
            ds.isel(azimuth=slice(missing, 180 - missing)),
            ds.isel(azimuth=slice(180 + missing, 360 - missing)),
        ],
        "azimuth",
        data_vars="minimal",
    )

    ds_out = util.remove_duplicate_rays(ds_in)
    ds_out = util.reindex_angle(
        ds_out,
        start_angle=start_ang,
        stop_angle=stop_ang,
        angle_res=1.0,
        direction=direction,
        method="nearest",
    )
    dsx = ds_out.copy(deep=True)
    dsy = dsx.pipe(util.ipol_time, a1gate_idx=a1gate, direction=direction)
    # see if all time values were reconstructed
    xr.testing.assert_equal(ds, dsy)

    # check emit warning
    if a1gate in [179, 180, 181] and missing and direction == 1:
        with pytest.warns(
            UserWarning, match="Rays might miss on beginning and/or end of sweep."
        ):
            dsx.pipe(util.ipol_time, direction=direction)


def test_get_sweep_keys():
    # Test finding sweep keys
    filename = DATASETS.fetch("sample_sgp_data.nc")
    dtree = io.open_cfradial1_datatree(filename)
    # set a fake group
    dtree["sneep_1"] = dtree["sweep_1"]
    keys = util.get_sweep_keys(dtree)
    assert keys == [
        "sweep_0",
        "sweep_1",
        "sweep_2",
        "sweep_3",
        "sweep_4",
        "sweep_5",
    ]


def test_apply_to_sweeps():
    # Fetch the sample radar file
    filename = DATASETS.fetch("sample_sgp_data.nc")

    # Open the radar file into a DataTree object
    dtree = io.open_cfradial1_datatree(filename)

    # Define a simple function to test with apply_to_sweeps
    def dummy_function(ds):
        """A dummy function that adds a constant field to the dataset."""
        ds["dummy_field"] = (
            ds["reflectivity_horizontal"] * 0
        )  # Adding a field with all zeros
        ds["dummy_field"].attrs = {"units": "dBZ", "long_name": "Dummy Field"}
        return ds

    # Apply the dummy function to all sweeps using apply_to_sweeps
    modified_dtree = util.apply_to_sweeps(dtree, dummy_function)

    # Verify that the dummy field has been added to each sweep
    sweep_keys = util.get_sweep_keys(modified_dtree)
    for key in sweep_keys:
        assert (
            "dummy_field" in modified_dtree[key].data_vars
        ), f"dummy_field not found in {key}"
        assert modified_dtree[key]["dummy_field"].attrs["units"] == "dBZ"
        assert modified_dtree[key]["dummy_field"].attrs["long_name"] == "Dummy Field"

    # Check that the original data has not been modified
    assert (
        "dummy_field" not in dtree["/"].data_vars
    ), "dummy_field should not be in the root node"

    # Test that an exception is raised when a function that causes an error is applied
    with pytest.raises(ValueError, match="This is an intentional error"):

        def error_function(ds):
            raise ValueError("This is an intentional error")

        util.apply_to_sweeps(dtree, error_function)


def test_apply_to_volume():
    # Fetch the sample radar file
    filename = DATASETS.fetch("sample_sgp_data.nc")

    # Open the radar file into a DataTree object
    dtree = io.open_cfradial1_datatree(filename)

    # Define a simple function to test with apply_to_volume
    def dummy_function(ds):
        """A dummy function that adds a constant field to the dataset."""
        ds["dummy_field"] = (
            ds["reflectivity_horizontal"] * 0
        )  # Adding a field with all zeros
        ds["dummy_field"].attrs = {"units": "dBZ", "long_name": "Dummy Field"}
        return ds

    # Apply the dummy function to all sweeps using apply_to_volume
    modified_dtree = util.apply_to_volume(dtree, dummy_function)

    # Verify that the modified_dtree is an instance of DataTree
    assert isinstance(
        modified_dtree, xr.DataTree
    ), "The result should be a DataTree instance."

    # Verify that the dummy field has been added to each sweep
    sweep_keys = util.get_sweep_keys(modified_dtree)
    for key in sweep_keys:
        assert (
            "dummy_field" in modified_dtree[key].data_vars
        ), f"dummy_field not found in {key}"
        assert modified_dtree[key]["dummy_field"].attrs["units"] == "dBZ"
        assert modified_dtree[key]["dummy_field"].attrs["long_name"] == "Dummy Field"

    # Check that the original DataTree (dtree) has not been modified
    original_sweep_keys = util.get_sweep_keys(dtree)
    for key in original_sweep_keys:
        assert (
            "dummy_field" not in dtree[key].data_vars
        ), f"dummy_field should not be in the original DataTree at {key}"

    # Test edge case: Apply a function that modifies only certain sweeps
    def selective_function(ds):
        """Only modifies sweeps with a specific condition."""
        if "reflectivity_horizontal" in ds:
            ds["selective_field"] = ds["reflectivity_horizontal"] * 1
        return ds

    # Apply the selective function to all sweeps using apply_to_volume
    selectively_modified_dtree = util.apply_to_volume(dtree, selective_function)

    # Verify that the selective field was added only where the condition was met
    for key in sweep_keys:
        if "reflectivity_horizontal" in modified_dtree[key].data_vars:
            assert (
                "selective_field" in selectively_modified_dtree[key].data_vars
            ), f"selective_field not found in {key} where it should have been added."
        else:
            assert (
                "selective_field" not in selectively_modified_dtree[key].data_vars
            ), f"selective_field should not be present in {key}"

    # Test that an exception is raised when a function that causes an error is applied
    with pytest.raises(ValueError, match="This is an intentional error"):

        def error_function(ds):
            raise ValueError("This is an intentional error")

        util.apply_to_volume(dtree, error_function)


def test_map_over_sweeps_decorator_dummy_function():
    """
    Test applying a dummy function to all sweep nodes using the map_over_sweeps decorator.
    """
    # Fetch the sample radar file
    filename = DATASETS.fetch("sample_sgp_data.nc")

    # Open the radar file into a DataTree object
    dtree = xd.io.open_cfradial1_datatree(filename)

    # Use the decorator on the dummy function
    @xd.map_over_sweeps
    def dummy_function(ds, refl="none"):
        ds = ds.assign(
            dummy_field=ds["reflectivity_horizontal"] * 0
        )  # Field with zeros
        ds["dummy_field"].attrs = {
            "unit": "dBZ",
            "long_name": "Dummy Field",
            "test": refl,
        }
        return ds

    # Apply using pipe and decorator
    dtree_modified = dtree.pipe(dummy_function, refl="test")

    # Check that the new field exists in sweep_0 and has the correct attributes
    sweep_0 = dtree_modified["sweep_0"]
    assert "dummy_field" in sweep_0.data_vars
    assert sweep_0.dummy_field.attrs["unit"] == "dBZ"
    assert sweep_0.dummy_field.attrs["long_name"] == "Dummy Field"
    assert sweep_0.dummy_field.attrs["test"] == "test"

    # Ensure all non-NaN values are 0 (accounting for -0.0 and NaN values)
    non_nan_values = np.nan_to_num(
        sweep_0.dummy_field.values
    )  # Convert NaNs to zero for comparison
    assert np.all(np.isclose(non_nan_values, 0))


def test_get_required_root_dataset():

    filename = DATASETS.fetch("cor-main131125105503.RAW2049")
    sweeps = [f"sweep_{i}" for i in range(10)]
    ls_ds = [xr.open_dataset(filename, engine="iris", group=sweep) for sweep in sweeps]
    root = _get_required_root_dataset(ls_ds, optional=True)
    elevations = [
        0.5,
        1.0,
        2.0,
        3.0,
        5.0,
        7.0,
        10.0,
        15.0,
        20.0,
        30.0,
    ]
    assert len(root.variables) == 10
    assert root.variables["time_coverage_start"] == "2013-11-25T10:55:04Z"
    assert root.variables["time_coverage_end"] == "2013-11-25T10:59:24Z"
    np.testing.assert_equal(
        root.variables["sweep_fixed_angle"].values, np.array(elevations)
    )
    assert len(list(root.attrs.keys())) == 10
    assert root.attrs["instrument_name"] == "Corozal, Radar"
    assert root.attrs["scan_name"] == "SURV_HV_300 "
    assert root.attrs["comment"] == "AEROCIVIL OPERATIONAL DUAL POLE SCAN"


def test_get_radar_calibration():
    filename = DATASETS.fetch("DWD-Vol-2_99999_20180601054047_00.h5")
    sweeps = [f"sweep_{i}" for i in range(10)]
    ls_ds = [xr.open_dataset(filename, engine="gamic", group=sweep) for sweep in sweeps]
    subgroup = _get_radar_calibration(ls_ds, model.radar_calibration_subgroup)
    assert len(subgroup.variables) == 6
    assert subgroup["noise_power_h"] == "-3.8298"
    assert subgroup["rx_loss_h"] == "3"
    assert subgroup["ant_gain_v"] == "43"
    assert subgroup["ant_gain_h"] == "43"


def test_get_subgroup():
    filename = DATASETS.fetch("71_20181220_060628.pvol.h5")
    sweeps = [f"sweep_{i}" for i in range(10)]
    ls_ds = [xr.open_dataset(filename, engine="odim", group=sweep) for sweep in sweeps]
    subgroup = _get_subgroup(ls_ds, model.radar_parameters_subgroup)
    assert len(subgroup.variables) == 3
    assert list(subgroup.variables) == ["longitude", "latitude", "altitude"]
    np.testing.assert_almost_equal(subgroup.longitude.values.item(), 151.20899963378906)
    np.testing.assert_almost_equal(subgroup.latitude.values.item(), -33.700801849365234)
    assert isinstance(subgroup.altitude.values.item(), float)



#!/usr/bin/env python
# Copyright (c) 2023-2024, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.

"""Tests for `io.backends.rainbow` module.

Ported from wradlib.
"""

import datetime
import zlib
from io import BytesIO

import numpy as np
import pytest
import xmltodict
from xarray import DataTree

from xradar.io.backends import open_rainbow_datatree, rainbow


def test_find_key():
    indict = {
        "A": {
            "AA": {"AAA": 0, "X": 1},
            "AB": {"ABA": 2, "X": 3},
            "AC": {"ACA": 4, "X": 5},
            "AD": [{"ADA": 4, "X": 2}],
        }
    }
    outdict = [
        {"X": 1, "AAA": 0},
        {"ABA": 2, "X": 3},
        {"X": 5, "ACA": 4},
        {"ADA": 4, "X": 2},
    ]

    assert list(rainbow.find_key("X", indict)) == outdict
    assert list(rainbow.find_key("Y", indict)) == []


def test_decompress():
    dstring = b"very special compressed string"
    cstring = zlib.compress(dstring)
    assert rainbow.decompress(cstring) == dstring


def test_get_rb_data_layout():
    assert rainbow.get_rb_data_layout(8) == (1, ">u1")
    assert rainbow.get_rb_data_layout(16) == (2, ">u2")
    assert rainbow.get_rb_data_layout(32) == (4, ">u4")
    with pytest.raises(ValueError):
        rainbow.get_rb_data_layout(128)


def test_get_rb_data_layout_big(monkeypatch):
    monkeypatch.setattr("sys.byteorder", "big")
    assert rainbow.get_rb_data_layout(8) == (1, "<u1")
    assert rainbow.get_rb_data_layout(16) == (2, "<u2")
    assert rainbow.get_rb_data_layout(32) == (4, "<u4")


def test_get_rb_data_attribute():
    data = xmltodict.parse(
        '<slicedata time="13:30:05" date="2013-04-26">'
        '#<rayinfo refid="startangle" blobid="0" '
        'rays="361" depth="16"/> '
        '#<rawdata blobid="1" rays="361" type="dBuZ" '
        'bins="400" min="-31.5" max="95.5" '
        'depth="8"/> #</slicedata>'
    )
    data = list(rainbow.find_key("@blobid", data))
    assert rainbow.get_rb_data_attribute(data[0], "blobid") == 0
    assert rainbow.get_rb_data_attribute(data[1], "blobid") == 1
    assert rainbow.get_rb_data_attribute(data[0], "rays") == 361
    assert rainbow.get_rb_data_attribute(data[1], "rays") == 361
    assert rainbow.get_rb_data_attribute(data[1], "bins") == 400
    with pytest.raises(KeyError):
        rainbow.get_rb_data_attribute(data[0], "Nonsense")
    assert rainbow.get_rb_data_attribute(data[0], "depth") == 16


def test_get_rb_blob_attribute():
    xmldict = xmltodict.parse('<BLOB blobid="0" size="737" compression="qt"></BLOB>')
    assert rainbow.get_rb_blob_attribute(xmldict, "compression") == "qt"
    assert rainbow.get_rb_blob_attribute(xmldict, "size") == "737"
    assert rainbow.get_rb_blob_attribute(xmldict, "blobid") == "0"
    with pytest.raises(KeyError):
        rainbow.get_rb_blob_attribute(xmldict, "Nonsense")


def test_get_rb_data_shape():
    data = xmltodict.parse(
        '<slicedata time="13:30:05" date="2013-04-26">'
        '#<rayinfo refid="startangle" blobid="0" '
        'rays="361" depth="16"/> #<rawdata blobid="1" '
        'rays="361" type="dBuZ" bins="400" '
        'min="-31.5" max="95.5" depth="8"/> #<flagmap '
        'blobid="2" rows="800" type="dBuZ" '
        'columns="400" min="-31.5" max="95.5" '
        'depth="6"/> #<defect blobid="3" type="dBuZ" '
        'columns="400" min="-31.5" max="95.5" '
        'depth="6"/> #<rawdata2 '
        'blobid="4" rows="800" type="dBuZ" '
        'columns="400" min="-31.5" max="95.5" '
        'depth="8"/> #</slicedata>'
    )
    data = list(rainbow.find_key("@blobid", data))
    assert rainbow.get_rb_data_shape(data[0]) == 361
    assert rainbow.get_rb_data_shape(data[1]) == (361, 400)
    assert rainbow.get_rb_data_shape(data[2]) == (800, 400, 6)
    assert rainbow.get_rb_data_shape(data[4]) == (800, 400)
    with pytest.raises(KeyError):
        rainbow.get_rb_data_shape(data[3])


def test_map_rb_data():
    indata = b"0123456789"
    outdata8 = np.array([48, 49, 50, 51, 52, 53, 54, 55, 56, 57], dtype=np.uint8)
    outdata16 = np.array([12337, 12851, 13365, 13879, 14393], dtype=np.uint16)
    outdata32 = np.array([808530483, 875902519], dtype=np.uint32)
    np.testing.assert_allclose(rainbow.map_rb_data(indata, 8), outdata8)
    np.testing.assert_allclose(rainbow.map_rb_data(indata, 16), outdata16)
    np.testing.assert_allclose(rainbow.map_rb_data(indata, 32), outdata32)
    flagdata = b"1"
    np.testing.assert_allclose(
        rainbow.map_rb_data(flagdata, 1, 8), [0, 0, 1, 1, 0, 0, 0, 1]
    )
    # added test for truncation
    np.testing.assert_allclose(rainbow.map_rb_data(flagdata, 1, 6), [0, 0, 1, 1, 0, 0])


def test_get_rb_blob_data():
    datastring = b'<BLOB blobid="0" size="737" compression="qt"></BLOB>'
    with pytest.raises(EOFError):
        rainbow.get_rb_blob_data(datastring, 1)


def test_get_rb_header():
    rb_header = (
        b'<volume version="5.34.16" '
        b'datetime="2013-07-03T08:33:55"'
        b' type="azi" owner="RainAnalyzer"> '
        b'<scan name="analyzer.azi" time="08:34:00" '
        b'date="2013-07-03">'
    )

    buf = BytesIO(rb_header)
    with pytest.raises(IOError):
        rainbow.get_rb_header(buf)


def test_get_rb_header_from_file(rainbow_file):
    with open(rainbow_file, "rb") as rb_fh:
        rb_header = rainbow.get_rb_header(rb_fh)
        assert rb_header["volume"]["@version"] == "5.36.5"


def test_rainbow_file_meta(rainbow_file):
    with rainbow.RainbowFile(rainbow_file, loaddata=False) as rbdict:
        assert rbdict.version == "5.36.5"
        assert rbdict.header
        assert rbdict.site_coords == (6.379967, 50.856633, 116.7)
        assert rbdict.first_dimension == "azimuth"
        assert rbdict.history is None
        assert rbdict.pargroup
        assert rbdict.sensorinfo == dict(
            [
                ("@id", "143DEX"),
                ("@name", "Gematronik"),
                ("@type", "gdrx"),
                ("alt", "116.700000"),
                ("beamwidth", "1.326"),
                ("lat", "50.856633"),
                ("lon", "6.379967"),
                ("wavelen", "0.0319"),
            ]
        )
        assert rbdict.datetime == datetime.datetime(2013, 5, 10, 0, 3, 17)
        assert rbdict.type == "vol"


def test_rainbow_file_data(rainbow_file):
    with rainbow.RainbowFile(rainbow_file, loaddata=True) as rbdict:
        sdata = rbdict.slices[0]["slicedata"]
        print(sdata["rayinfo"]["data"][0])
        np.testing.assert_equal(sdata["rayinfo"]["data"][0], np.array(47.0159912109375))
        np.testing.assert_equal(sdata["rawdata"]["data"][0, 0], np.array(113))
        assert sdata["rawdata"]["data"].shape == (361, 400)


def test_open_rainbow_datatree(rainbow_file):
    # Define kwargs to pass into the function
    kwargs = {
        "sweep": [0, 1],  # Specify sweeps to extract
        "first_dim": "auto",
        "reindex_angle": False,
        "site_coords": True,
    }

    # Call the function with the Rainbow file
    dtree = open_rainbow_datatree(rainbow_file, **kwargs)

    # Assertions to check DataTree structure
    assert isinstance(dtree, DataTree), "Expected a DataTree instance"
    assert "/" in dtree.subtree, "Root group should be present in the DataTree"
    assert (
        "/radar_parameters" in dtree.subtree
    ), "Radar parameters group should be in the DataTree"
    assert (
        "/georeferencing_correction" in dtree.subtree
    ), "Georeferencing correction group should be in the DataTree"
    assert (
        "/radar_calibration" in dtree.subtree
    ), "Radar calibration group should be in the DataTree"

    # Check if the specified sweep groups are attached
    sweep_groups = [key for key in dtree.match("sweep_*")]
    assert len(sweep_groups) == 2, "Expected two sweep groups in the DataTree"
    sample_sweep = sweep_groups[0]

    # Verify the presence of specific data variables
    assert (
        "DBZH" in dtree[sample_sweep].variables.keys()
    ), "Expected 'DBZH' variable in the sweep group"
    assert dtree[sample_sweep]["DBZH"].shape == (
        361,
        400,
    ), "Shape mismatch for 'DBZH' variable"

    # Validate coordinates are attached correctly in the root dataset
    assert (
        "latitude" in dtree[sample_sweep]
    ), "Latitude should be attached to the root dataset"
    assert (
        "longitude" in dtree[sample_sweep]
    ), "Longitude should be attached to the root dataset"
    assert (
        "altitude" in dtree[sample_sweep]
    ), "Altitude should be attached to the root dataset"

    # Validate attributes
    assert len(dtree.attrs) == 9



#!/usr/bin/env python
# Copyright (c) 2023-2024, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.

"""
CfRadial1 Output
================

This sub-module contains the writer for export of CfRadial1-based radar
data.

Example::

    import xradar as xd
    xd.io.to_cfradial1(dtree, filename, calibs=True)

.. autosummary::
   :nosignatures:
   :toctree: generated/

   {}
"""

__all__ = [
    "to_cfradial1",
]

__doc__ = __doc__.format("\n   ".join(__all__))

from importlib.metadata import version

import numpy as np
import xarray as xr


def _calib_mapper(calib_params):
    """
    Map calibration parameters to a new dataset format.

    Parameters
    ----------
    calib_params: xarray.Dataset
        Calibration parameters dataset.

    Returns
    -------
    xarray.Dataset
        New dataset with mapped calibration parameters.
    """
    new_data_vars = {}
    for var in calib_params.data_vars:
        data_array = calib_params[var]
        new_data_vars["r_calib_" + var] = xr.DataArray(
            data=data_array.data[np.newaxis, ...],
            dims=["r_calib"] + list(data_array.dims),
            coords={"r_calib": [0]},
            attrs=data_array.attrs,
        )
    radar_calib_renamed = xr.Dataset(new_data_vars)
    radar_calib_renamed = radar_calib_renamed.drop_vars("r_calib", errors="ignore")
    return radar_calib_renamed


def _main_info_mapper(dtree):
    """
    Map main radar information from a radar xarray dataset.

    Parameters
    ----------
    dtree: xarray.DataTree
        Radar xarray.DataTree.

    Returns
    -------
    xarray.Dataset
        Dataset containing the mapped radar information.
    """
    dataset = dtree.root.to_dataset()
    dataset = dataset.drop_vars("sweep_group_name", errors="ignore")

    # Rename "sweep_fixed_angle" to "fixed_angle" if it exists
    if "sweep_fixed_angle" in dataset:
        dataset = dataset.rename({"sweep_fixed_angle": "fixed_angle"})

    return dataset


def _variable_mapper(dtree, dim0=None):
    """
    Map radar variables for different sweep groups.

    Parameters
    ----------
    dtree: xarray.DataTree
        Radar xarray.DataTree.
    dim0: str
        Either `azimuth` or `elevation`

    Returns
    -------
    xarray.Dataset
        Dataset containing mapped radar variables.
    """

    sweep_info = _sweep_info_mapper(dtree)
    vol_info = _main_info_mapper(dtree)
    if "fixed_angle" in vol_info:
        vol_info = vol_info.drop_vars("fixed_angle")
    sweep_datasets = []
    for grp in dtree.groups:
        if "sweep" in grp:
            data = dtree[grp].to_dataset()

            # handling first dimension
            if dim0 is None:
                dim0 = (
                    "elevation"
                    if data.sweep_mode.load().astype(str) == "rhi"
                    else "azimuth"
                )
                if dim0 not in data.dims:
                    dim0 = "time"
                    assert dim0 in data.dims

            # swap dims, if needed
            if dim0 != "time" and dim0 in data.dims:
                data = data.swap_dims({dim0: "time"})

            # sort in any case
            data = data.sortby("time")

            data = data.drop_vars(["x", "y", "z"], errors="ignore")

            # Convert to a dataset and append to the list
            sweep_datasets.append(data)

    # need to use combine_by_coords to correctly test for
    # incompatible attrs on DataArray's
    result_dataset = xr.combine_by_coords(
        sweep_datasets,
        data_vars="all",
        compat="no_conflicts",
        join="outer",
        coords="minimal",
        combine_attrs="no_conflicts",
    )

    drop_variables = [
        "sweep_fixed_angle",
        "sweep_number",
        "sweep_mode",
        "prt_mode",
        "follow_mode",
    ]
    result_dataset = result_dataset.drop_vars(drop_variables, errors="ignore")

    drop_coords = ["latitude", "longitude", "altitude", "spatial_ref", "crs_wkt"]
    result_dataset = result_dataset.drop_vars(drop_coords, errors="ignore")

    result_dataset = result_dataset.update(sweep_info)
    sweep_indices = calculate_sweep_indices(dtree, result_dataset)
    result_dataset = result_dataset.update(sweep_indices)
    result_dataset = result_dataset.reset_coords(["elevation", "azimuth"])
    result_dataset = result_dataset.update(vol_info)
    return result_dataset


def _sweep_info_mapper(dtree):
    """
    Extract specified sweep information variables from a radar xarray.DataTree

    Parameters
    ----------
    dtree: xarray.DataTree
        Radar xarray.DataTree.

    Returns
    -------
    xarray.Dataset
        Dataset containing the specified sweep information variables.
    """
    dataset = xr.Dataset()

    sweep_vars = [
        "sweep_number",
        "sweep_mode",
        "polarization_mode",
        "prt_mode",
        "follow_mode",
        "sweep_fixed_angle",
        "sweep_start_ray_index",
        "sweep_end_ray_index",
    ]

    for var_name in sweep_vars:
        var_data_list = [
            (
                np.unique(dtree[s][var_name].values[np.newaxis, ...])
                if var_name in dtree[s]
                else np.array([np.nan])
            )
            for s in dtree.groups
            if "sweep" in s
        ]

        var_attrs_list = [
            dtree[s][var_name].attrs if var_name in dtree[s] else {}
            for s in dtree.groups
            if "sweep" in s
        ]

        if not var_data_list:
            var_data = np.array([np.nan])
        else:
            var_data = np.concatenate(var_data_list)

        var_attrs = {}
        for attrs in var_attrs_list:
            var_attrs.update(attrs)

        var_data_array = xr.DataArray(var_data, dims=("sweep",), attrs=var_attrs)
        dataset[var_name] = var_data_array

    dataset = dataset.rename({"sweep_fixed_angle": "fixed_angle"})

    return dataset


def calculate_sweep_indices(dtree, dataset=None):
    """
    Calculate sweep start and end ray indices for elevation
    values in a radar dataset.

    Parameters
    ----------
    dtree: xarray.DataTree
        Radar xarray.DataTree containing elevation values for different sweep groups.
    dataset: xarray.Dataset, optional
        An optional dataset to which the calculated indices will be added.
        If None, a new dataset will be created.

    Returns:
    xarray.Dataset
        Dataset with sweep start and end ray indices.
    """
    if dataset is None:
        dataset = xr.Dataset()

    sweep_start_ray_index = []
    sweep_end_ray_index = []

    cumulative_size = 0

    try:
        for group_name in dtree.groups:
            if "sweep" in group_name:
                ele_size = dtree[group_name].elevation.size
                sweep_start_ray_index.append(cumulative_size)
                sweep_end_ray_index.append(cumulative_size + ele_size - 1)
                cumulative_size += ele_size

    except KeyError as e:
        print(
            f"Error: The sweep group '{e.args[0]}' was not found in radar datatree. Skipping..."
        )

    dataset["sweep_start_ray_index"] = xr.DataArray(
        sweep_start_ray_index,
        dims=("sweep",),
        attrs={"standard_name": "index_of_first_ray_in_sweep"},
    )

    dataset["sweep_end_ray_index"] = xr.DataArray(
        sweep_end_ray_index,
        dims=("sweep",),
        attrs={"standard_name": "index_of_last_ray_in_sweep"},
    )

    return dataset


def to_cfradial1(dtree=None, filename=None, calibs=True):
    """
    Convert a radar xarray.DataTree to the CFRadial1 format
    and save it to a file. Ensure that the resulting dataset
    is well-formed and does not include specified extraneous variables.

    Parameters
    ----------
    dtree: xarray.DataTree
        Radar xarary.DataTree object.
    filename: str, optional
        The name of the output netCDF file.
    calibs: Bool, optional
        Whether to include calibration parameters.
    """
    # Generate the initial ds_cf using the existing mapping functions
    dataset = _variable_mapper(dtree)

    # Handle calibration parameters
    if calibs:
        if "radar_calibration" in dtree:
            calib_params = dtree["radar_calibration"].to_dataset()
            calibs = _calib_mapper(calib_params)
            dataset.update(calibs)

    # Add additional parameters if they exist in dtree
    if "radar_parameters" in dtree:
        radar_params = dtree["radar_parameters"].to_dataset().reset_coords()
        dataset.update(radar_params)

    if "georeferencing_correction" in dtree:
        radar_georef = dtree["georeferencing_correction"].to_dataset().reset_coords()
        dataset.update(radar_georef)

    # Ensure that the data type of sweep_mode and similar variables matches
    if "sweep_mode" in dataset.variables:
        dataset["sweep_mode"] = dataset["sweep_mode"].astype("S")

    # Update global attributes
    dataset.attrs = dtree.attrs
    dataset.attrs["Conventions"] = "Cf/Radial"
    dataset.attrs["version"] = "1.2"
    xradar_version = version("xradar")
    dataset.attrs["history"] += f": xradar v{xradar_version} CfRadial1 export"

    if filename is None:
        time = str(dataset.time[0].dt.strftime("%Y%m%d_%H%M%S").values)
        filename = f"cfrad1_{dataset.instrument_name}_{time}.nc"

    dataset.to_netcdf(filename, format="netcdf4")



#!/usr/bin/env python
# Copyright (c) 2024, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.

import pytest
import xarray as xr
from xarray import MergeError

import xradar as xd


def test_to_cfradial1(cfradial1_file):
    """Test the conversion from DataTree to CfRadial1 format."""
    with xd.io.open_cfradial1_datatree(cfradial1_file) as dtree:

        # Call the conversion function
        ds_cf1 = xd.transform.to_cfradial1(dtree)

        # Verify key attributes and data structures in the resulting dataset
        assert isinstance(ds_cf1, xr.Dataset), "Output is not a valid xarray Dataset"
        assert (
            "Conventions" in ds_cf1.attrs and ds_cf1.attrs["Conventions"] == "Cf/Radial"
        )
        assert (
            "sweep_mode" in ds_cf1.variables
        ), "Missing sweep_mode in converted dataset"
        assert ds_cf1.attrs["version"] == "1.2", "Incorrect CfRadial version"


def test_to_cfradial2(cfradial1_file):
    """Test the conversion from CfRadial1 to CfRadial2 DataTree format."""
    with xd.io.open_cfradial1_datatree(cfradial1_file) as dtree:

        # Convert to CfRadial1 dataset first
        ds_cf1 = xd.transform.to_cfradial1(dtree)

        # Call the conversion back to CfRadial2
        dtree_cf2 = xd.transform.to_cfradial2(ds_cf1)

        # Verify key attributes and data structures in the resulting datatree
        assert isinstance(dtree_cf2, xr.DataTree), "Output is not a valid DataTree"
        assert "radar_parameters" in dtree_cf2, "Missing radar_parameters in DataTree"
        assert dtree_cf2.attrs == ds_cf1.attrs, "Attributes mismatch between formats"


def test_to_cfradial1_with_different_range_shapes(nexradlevel2_bzfile):
    with xd.io.open_nexradlevel2_datatree(nexradlevel2_bzfile) as dtree:
        ds_cf1 = xd.transform.to_cfradial1(dtree)
        # Verify key attributes and data structures in the resulting dataset
        assert isinstance(ds_cf1, xr.Dataset), "Output is not a valid xarray Dataset"
        assert (
            "Conventions" in ds_cf1.attrs and ds_cf1.attrs["Conventions"] == "Cf/Radial"
        )
        assert (
            "sweep_mode" in ds_cf1.variables
        ), "Missing sweep_mode in converted dataset"
        assert ds_cf1.attrs["version"] == "1.2", "Incorrect CfRadial version"
        assert ds_cf1.sizes.mapping == {"time": 5400, "range": 1832, "sweep": 11}

        # Call the conversion back to CfRadial2
        dtree_cf2 = xd.transform.to_cfradial2(ds_cf1)
        # Verify key attributes and data structures in the resulting datatree
        assert isinstance(dtree_cf2, xr.DataTree), "Output is not a valid DataTree"
        # todo: this needs to be fixed in nexrad level2reader
        # assert "radar_parameters" in dtree_cf2, "Missing radar_parameters in DataTree"
        assert dtree_cf2.attrs == ds_cf1.attrs, "Attributes mismatch between formats"


def test_to_cfradial1_error_with_different_range_bin_sizes(gamic_file):
    with xd.io.open_gamic_datatree(gamic_file) as dtree:
        with pytest.raises(MergeError):
            xd.transform.to_cfradial1(dtree)



#!/usr/bin/env python
# Copyright (c) 2022-2025, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.

"""Tests for `io` module."""

import io
import tempfile

import fsspec
import h5py
import numpy as np
import pytest
import xarray as xr

import xradar.io
from tests import skip_import
from xradar.io import (
    open_cfradial1_datatree,
    open_datamet_datatree,
    open_gamic_datatree,
    open_iris_datatree,
    open_nexradlevel2_datatree,
    open_odim_datatree,
    open_rainbow_datatree,
)
from xradar.model import (
    non_standard_sweep_dataset_vars,
    required_sweep_metadata_vars,
    sweep_dataset_vars,
)


def test_open_cfradial1_datatree(cfradial1_file):
    dtree = open_cfradial1_datatree(cfradial1_file, first_dim="time", site_coords=False)
    attrs = dtree.attrs

    # root_attrs
    assert (
        attrs["Conventions"]
        == "CF/Radial instrument_parameters radar_parameters radar_calibration geometry_correction"
    )
    assert attrs["version"] == "1.2"
    assert attrs["title"] == "TIMREX"
    assert attrs["instrument_name"] == "SPOLRVP8"
    assert attrs["platform_is_mobile"] == "false"

    # root vars
    rvars = dtree.data_vars
    assert rvars["volume_number"] == 36
    assert rvars["platform_type"] == b"fixed"
    assert rvars["instrument_type"] == b"radar"
    assert rvars["time_coverage_start"] == b"2008-06-04T00:15:03Z"
    assert rvars["time_coverage_end"] == b"2008-06-04T00:22:17Z"
    np.testing.assert_almost_equal(rvars["latitude"].values, np.array(22.526699))
    np.testing.assert_almost_equal(rvars["longitude"].values, np.array(120.4335022))
    np.testing.assert_almost_equal(rvars["altitude"].values, np.array(45.0000018))

    # iterate over subgroups and check some values
    moments = ["DBZ", "VR"]
    elevations = [0.5, 1.1, 1.8, 2.6, 3.6, 4.7, 6.5, 9.1, 12.8]
    azimuths = [483, 483, 482, 483, 481, 482, 482, 484, 483]
    ranges = [996, 996, 996, 996, 996, 996, 996, 996, 996]
    for grp in dtree.groups:
        # only iterate sweep groups
        if "sweep" not in grp:
            continue
        ds = dtree[grp].ds
        i = ds.sweep_number.values
        assert i == int(grp[7:])
        assert dict(ds.sizes) == {"time": azimuths[i], "range": ranges[i]}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == set(moments)
        assert set(ds.data_vars) & (required_sweep_metadata_vars) == set(
            required_sweep_metadata_vars ^ {"azimuth", "elevation"}
        )
        assert set(ds.coords) == {
            "azimuth",
            "elevation",
            "time",
            "range",
        }
        assert np.round(ds.sweep_fixed_angle.values.item(), 1) == elevations[i]
        assert ds.sweep_mode == "azimuth_surveillance"


def test_open_cfradial1_dataset(cfradial1_file):
    # open first sweep group
    with xr.open_dataset(cfradial1_file, group="sweep_0", engine="cfradial1") as ds:
        assert list(ds.dims) == ["azimuth", "range"]
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == {"DBZ", "VR"}
        assert ds.sweep_number == 0

    # open last sweep group
    with xr.open_dataset(cfradial1_file, group="sweep_8", engine="cfradial1") as ds:
        assert list(ds.dims) == ["azimuth", "range"]
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == {"DBZ", "VR"}
        assert ds.sweep_number == 8


@pytest.mark.parametrize("sweep", ["sweep_0", 0, [0, 1], ["sweep_0", "sweep_1"]])
def test_open_odim_datatree_sweep(odim_file, sweep):
    dtree = open_odim_datatree(odim_file, sweep=sweep)
    if isinstance(sweep, (str, int)):
        lswp = len([sweep])
    else:
        lswp = len(sweep)
    assert len(dtree.match("sweep_*")) == lswp


def test_open_odim_datatree(odim_file):
    dtree = open_odim_datatree(odim_file)
    # root_attrs
    attrs = dtree.attrs
    assert attrs["Conventions"] == "ODIM_H5/V2_2"

    # root vars
    rvars = dtree.data_vars
    assert rvars["volume_number"] == 0
    assert rvars["platform_type"] == "fixed"
    assert rvars["instrument_type"] == "radar"
    assert rvars["time_coverage_start"] == "2018-12-20T06:06:28Z"
    assert rvars["time_coverage_end"] == "2018-12-20T06:10:42Z"
    np.testing.assert_almost_equal(rvars["latitude"].values, np.array(-33.7008018))
    np.testing.assert_almost_equal(rvars["longitude"].values, np.array(151.2089996))
    np.testing.assert_almost_equal(rvars["altitude"].values, np.array(195.0))

    # iterate over subgroups and check some values
    moments = ["PHIDP", "VRADH", "DBZH", "TH", "ZDR", "RHOHV", "WRADH", "KDP"]
    elevations = [
        0.5,
        0.9,
        1.3,
        1.8,
        2.4,
        3.1,
        4.2,
        5.6,
        7.4,
        10.0,
        13.3,
        17.9,
        23.9,
        32.0,
    ]
    azimuths = [360] * 14
    ranges = [
        1200,
        1200,
        1200,
        1200,
        1200,
        1180,
        1100,
        1100,
        590,
        480,
        360,
        280,
        200,
        200,
    ]
    for i, grp in enumerate(dtree.match("sweep_*")):
        ds = dtree[grp].ds
        assert dict(ds.sizes) == {"azimuth": azimuths[i], "range": ranges[i]}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == set(moments)
        assert set(ds.data_vars) & (required_sweep_metadata_vars) == set(
            required_sweep_metadata_vars ^ {"azimuth", "elevation"}
        )
        assert set(ds.coords) == {
            "azimuth",
            "elevation",
            "time",
            "latitude",
            "longitude",
            "altitude",
            "range",
        }
        assert np.round(ds.elevation.mean().values.item(), 1) == elevations[i]
        assert ds.sweep_number.values == int(grp[6:])


@pytest.mark.parametrize("first_dim", ["auto", "time"])
@pytest.mark.parametrize("fix_second_angle", [False, True])
def test_open_odim_dataset(odim_file, first_dim, fix_second_angle):
    # open first sweep group
    with xr.open_dataset(
        odim_file,
        group="sweep_0",
        engine="odim",
        first_dim=first_dim,
        fix_second_angle=fix_second_angle,
    ) as ds:
        dim0 = "time" if first_dim == "time" else "azimuth"
        assert dict(ds.sizes) == {dim0: 360, "range": 1200}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == {"WRADH", "VRADH", "PHIDP", "DBZH", "RHOHV", "KDP", "TH", "ZDR"}
        assert ds.sweep_number == 0

    # open last sweep group
    with xr.open_dataset(
        odim_file,
        group="sweep_11",
        engine="odim",
        first_dim=first_dim,
        fix_second_angle=fix_second_angle,
    ) as ds:
        assert dict(ds.sizes) == {dim0: 360, "range": 280}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == {"VRADH", "KDP", "WRADH", "TH", "RHOHV", "PHIDP", "ZDR", "DBZH"}
        assert ds.sweep_number == 11


def test_open_odim_dataset_stream(odim_file):
    with open(odim_file, mode="rb") as fhandle:
        contents = io.BytesIO(fhandle.read())
        with xr.open_dataset(contents, group="sweep_0", engine="odim") as ds:
            assert isinstance(ds, xr.Dataset)


def test_open_odim_dataset_fsspec(odim_file):
    with fsspec.open(odim_file, mode="rb") as fhandle:
        with xr.open_dataset(fhandle, group="sweep_0", engine="odim") as ds:
            assert isinstance(ds, xr.Dataset)


def test_open_odim_store(odim_file):
    store = xradar.io.backends.odim.OdimStore.open(odim_file, group="sweep_0")
    assert store.substore[0].root.a1gate == 86
    assert store.substore[0].root.site_coords == (
        151.20899963378906,
        -33.700801849365234,
        195.0,
    )
    assert store.substore[0].root._get_site_coords() == (
        151.20899963378906,
        -33.700801849365234,
        195.0,
    )
    assert store.substore[0].root.sweep_fixed_angle == 0.5
    assert store.substore[0].root._get_time() == np.datetime64(
        "2018-12-20T06:06:28", "s"
    )


@pytest.mark.parametrize("sweep", ["sweep_0", 0, [0, 1], ["sweep_0", "sweep_1"]])
def test_open_gamic_datatree_sweep(gamic_file, sweep):
    dtree = open_gamic_datatree(gamic_file, sweep=sweep)
    if isinstance(sweep, (str, int)):
        lswp = len([sweep])
    else:
        lswp = len(sweep)
    assert len(dtree.match("sweep_*")) == lswp


def test_open_gamic_datatree(gamic_file):
    dtree = open_gamic_datatree(gamic_file)

    # root_attrs
    attrs = dtree.attrs
    assert attrs["Conventions"] == "None"

    # root vars
    rvars = dtree.data_vars
    assert rvars["volume_number"] == 0
    assert rvars["platform_type"] == "fixed"
    assert rvars["instrument_type"] == "radar"
    assert rvars["time_coverage_start"] == "2018-06-01T05:40:47Z"
    assert rvars["time_coverage_end"] == "2018-06-01T05:44:16Z"
    np.testing.assert_almost_equal(rvars["latitude"].values, np.array(50.9287272))
    np.testing.assert_almost_equal(rvars["longitude"].values, np.array(6.4569489))
    np.testing.assert_almost_equal(rvars["altitude"].values, np.array(310.0))

    # iterate over subgroups and check some values
    moments = [
        "WRADH",
        "WRADV",
        "VRADH",
        "VRADV",
        "PHIDP",
        "DBTH",
        "DBTV",
        "DBZH",
        "DBZV",
        "RHOHV",
        "KDP",
        "ZDR",
    ]
    elevations = [
        28.0,
        18.0,
        14.0,
        11.0,
        8.2,
        6.0,
        4.5,
        3.1,
        1.7,
        0.6,
    ]
    azimuths = [361, 361, 361, 360, 361, 360, 360, 361, 360, 360]
    ranges = [
        360,
        500,
        620,
        800,
        1050,
        1400,
        1000,
        1000,
        1000,
        1000,
    ]
    for i, grp in enumerate(dtree.match("sweep_*")):
        ds = dtree[grp].ds
        assert dict(ds.sizes) == {"azimuth": azimuths[i], "range": ranges[i]}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == set(moments)
        assert set(ds.data_vars) & (required_sweep_metadata_vars) == set(
            required_sweep_metadata_vars ^ {"azimuth", "elevation"}
        )
        assert set(ds.coords) == {
            "azimuth",
            "elevation",
            "time",
            "latitude",
            "longitude",
            "altitude",
            "range",
        }
        assert np.round(ds.elevation.mean().values.item(), 1) == elevations[i]
        assert ds.sweep_number == i


@pytest.mark.parametrize("first_dim", ["auto", "time"])
@pytest.mark.parametrize("fix_second_angle", [False, True])
def test_open_gamic_dataset(gamic_file, first_dim, fix_second_angle):
    # open first sweep group
    with xr.open_dataset(
        gamic_file,
        group="sweep_0",
        engine="gamic",
        first_dim=first_dim,
        fix_second_angle=fix_second_angle,
    ) as ds:
        dim0 = "time" if first_dim == "time" else "azimuth"
        assert dict(ds.sizes) == {dim0: 361, "range": 360}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == {
            "WRADH",
            "WRADV",
            "VRADH",
            "VRADV",
            "PHIDP",
            "DBTH",
            "DBTV",
            "DBZH",
            "DBZV",
            "RHOHV",
            "KDP",
            "ZDR",
        }
        assert ds.sweep_number == 0

    # open last sweep group
    with xr.open_dataset(
        gamic_file,
        group="sweep_9",
        engine="gamic",
        first_dim=first_dim,
        fix_second_angle=fix_second_angle,
    ) as ds:
        assert dict(ds.sizes) == {dim0: 360, "range": 1000}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == {
            "WRADH",
            "WRADV",
            "VRADH",
            "VRADV",
            "PHIDP",
            "DBTH",
            "DBTV",
            "DBZH",
            "DBZV",
            "RHOHV",
            "KDP",
            "ZDR",
        }
        assert ds.sweep_number == 9


def test_open_gamic_dataset_stream(gamic_file):
    with open(gamic_file, mode="rb") as fhandle:
        contents = io.BytesIO(fhandle.read())
        with xr.open_dataset(contents, group="sweep_9", engine="gamic") as ds:
            assert isinstance(ds, xr.Dataset)
            print(ds)


def test_open_gamic_dataset_fsspec(gamic_file):
    with fsspec.open(gamic_file, mode="rb") as fhandle:
        with xr.open_dataset(fhandle, group="sweep_9", engine="gamic") as ds:
            assert isinstance(ds, xr.Dataset)
            print(ds)


def test_open_gamic_store(gamic_file):
    store = xradar.io.backends.gamic.GamicStore.open(gamic_file, group="sweep_0")
    assert store.root.site_coords == (6.4569489, 50.9287272, 310.0)
    assert store.root._get_site_coords() == (6.4569489, 50.9287272, 310.0)
    assert store.root.sweep_fixed_angle == 28.0
    assert store.root._get_time() == np.datetime64("2018-06-01T05:40:47.041000", "us")


def test_open_gamic_dataset_reindex(gamic_file):
    # open first sweep group
    reindex_angle = dict(start_angle=0, stop_angle=360, angle_res=1.0, direction=1)
    with xr.open_dataset(
        gamic_file, group="sweep_0", engine="gamic", reindex_angle=reindex_angle
    ) as ds:
        assert dict(ds.sizes) == {"azimuth": 360, "range": 360}


def test_open_furuno_scn_dataset(furuno_scn_file):
    # open sweep group
    with xr.open_dataset(furuno_scn_file, first_dim="time", engine="furuno") as ds:
        assert dict(ds.sizes) == {"time": 1376, "range": 602}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == {"KDP", "VRADH", "ZDR", "DBZH", "WRADH", "RHOHV", "PHIDP"}
        for key, value in ds.data_vars.items():
            if key in [
                "RATE",
                "KDP",
                "VRADH",
                "ZDR",
                "DBZH",
                "WRADH",
                "RHOHV",
                "PHIDP",
            ]:
                assert value.encoding["_FillValue"] == 0.0
            elif key in ["azimuth", "elevation"]:
                assert value.encoding["_FillValue"] == np.ma.minimum_fill_value(
                    value.encoding["dtype"]
                )
            else:
                assert value.encoding.get("_FillValue", None) is None
        assert ds.sweep_number == 0

    # open sweep group, auto
    with xr.open_dataset(
        furuno_scn_file,
        engine="furuno",
    ) as ds:
        assert dict(ds.sizes) == {"azimuth": 1376, "range": 602}
        assert ds.sweep_number == 0


def test_open_furuno_scnx_dataset(furuno_scnx_file):
    # open sweep group
    with xr.open_dataset(furuno_scnx_file, first_dim="time", engine="furuno") as ds:
        assert dict(ds.sizes) == {"time": 722, "range": 936}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == {"KDP", "VRADH", "ZDR", "DBZH", "WRADH", "RHOHV", "PHIDP"}

        for key, value in ds.data_vars.items():
            if key in [
                "RATE",
                "KDP",
                "VRADH",
                "ZDR",
                "DBZH",
                "WRADH",
                "RHOHV",
                "PHIDP",
            ]:
                assert value.encoding["_FillValue"] == 0.0
            elif key in ["azimuth", "elevation"]:
                assert value.encoding["_FillValue"] == np.ma.minimum_fill_value(
                    value.encoding["dtype"]
                )
            else:
                assert value.encoding.get("_FillValue", None) is None
        assert ds.sweep_number == 0

    # open sweep group, auto
    with xr.open_dataset(
        furuno_scnx_file,
        engine="furuno",
    ) as ds:
        assert dict(ds.sizes) == {"azimuth": 722, "range": 936}
        assert ds.sweep_number == 0


def test_open_rainbow_datatree(rainbow_file):
    dtree = open_rainbow_datatree(rainbow_file)

    # root_attrs
    attrs = dtree.attrs
    assert attrs["Conventions"] == "None"

    # root vars
    rvars = dtree.data_vars
    assert rvars["volume_number"] == 0
    assert rvars["platform_type"] == "fixed"
    assert rvars["instrument_type"] == "radar"
    assert rvars["time_coverage_start"] == "2013-05-10T00:00:06Z"
    assert rvars["time_coverage_end"] == "2013-05-10T00:03:14Z"
    np.testing.assert_almost_equal(rvars["latitude"].values, np.array(50.856633))
    np.testing.assert_almost_equal(rvars["longitude"].values, np.array(6.379967))
    np.testing.assert_almost_equal(rvars["altitude"].values, np.array(116.7))

    # iterate over subgroups and check some values
    moments = [
        "DBZH",
    ]
    elevations = [
        0.6,
        1.4,
        2.4,
        3.5,
        4.8,
        6.3,
        8.0,
        9.9,
        12.2,
        14.8,
        17.9,
        21.3,
        25.4,
        30.0,
    ]
    azimuths = [361] * 14
    ranges = [400] * 14
    for i, grp in enumerate(dtree.match("sweep_*")):
        ds = dtree[grp].ds
        assert dict(ds.sizes) == {"azimuth": azimuths[i], "range": ranges[i]}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == set(moments)
        assert set(ds.data_vars) & (required_sweep_metadata_vars) == set(
            required_sweep_metadata_vars ^ {"azimuth", "elevation"}
        )
        assert set(ds.coords) == {
            "azimuth",
            "elevation",
            "time",
            "latitude",
            "longitude",
            "altitude",
            "range",
        }
        assert np.round(ds.elevation.mean().values.item(), 1) == elevations[i]
        assert ds.sweep_number == i


def test_open_rainbow_dataset(rainbow_file):
    # open first sweep group
    with xr.open_dataset(rainbow_file, group="sweep_0", engine="rainbow") as ds:
        assert dict(ds.sizes) == {"azimuth": 361, "range": 400}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == {
            "DBZH",
        }
        assert ds.sweep_number == 0

    # open last sweep group
    with xr.open_dataset(rainbow_file, group="sweep_13", engine="rainbow") as ds:
        assert dict(ds.sizes) == {"azimuth": 361, "range": 400}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == {
            "DBZH",
        }
        assert ds.sweep_number == 13

    # open last sweep group, auto
    with xr.open_dataset(
        rainbow_file,
        group="sweep_13",
        engine="rainbow",
        backend_kwargs=dict(first_dim="time"),
    ) as ds:
        assert dict(ds.sizes) == {"time": 361, "range": 400}
        assert ds.sweep_number == 13


def test_open_iris_datatree(iris0_file):
    dtree = open_iris_datatree(iris0_file)

    # root_attrs
    attrs = dtree.attrs
    assert attrs["Conventions"] == "None"

    # root vars
    rvars = dtree.data_vars
    assert rvars["volume_number"] == 0
    assert rvars["platform_type"] == "fixed"
    assert rvars["instrument_type"] == "radar"
    assert rvars["time_coverage_start"] == "2013-11-25T10:55:04Z"
    assert rvars["time_coverage_end"] == "2013-11-25T10:59:24Z"
    np.testing.assert_almost_equal(rvars["latitude"].values, np.array(9.331))
    np.testing.assert_almost_equal(rvars["longitude"].values, np.array(-75.2829999))
    np.testing.assert_almost_equal(rvars["altitude"].values, np.array(143.0))

    # iterate over subgroups and check some values
    moments = [
        "ZDR",
        "RHOHV",
        "DBZH",
        "PHIDP",
        "KDP",
        "VRADH",
    ]
    elevations = [
        0.5,
        1.0,
        2.0,
        3.0,
        5.0,
        7.0,
        10.0,
        15.0,
        20.0,
        30.0,
    ]
    azimuths = [360] * 10
    ranges = [664] * 10
    i = 0
    for grp in dtree.match("sweep_*"):
        ds = dtree[grp].ds
        assert dict(ds.sizes) == {"azimuth": azimuths[i], "range": ranges[i]}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == set(moments)
        assert set(ds.data_vars) & (required_sweep_metadata_vars) == set(
            required_sweep_metadata_vars ^ {"azimuth", "elevation"}
        )
        assert set(ds.coords) == {
            "azimuth",
            "elevation",
            "time",
            "latitude",
            "longitude",
            "altitude",
            "range",
        }
        assert np.round(ds.elevation.mean().values.item(), 1) == elevations[i]
        assert ds.sweep_number == i
        i += 1


def test_open_iris0_dataset(iris0_file):
    # open first sweep group
    with xr.open_dataset(iris0_file, group="sweep_0", engine="iris") as ds:
        assert dict(ds.sizes) == {"azimuth": 360, "range": 664}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == {
            "DBZH",
            "VRADH",
            "KDP",
            "RHOHV",
            "PHIDP",
            "ZDR",
        }
        assert ds.sweep_number == 0

    # open last sweep group
    with xr.open_dataset(iris0_file, group="sweep_9", engine="iris") as ds:
        assert dict(ds.sizes) == {"azimuth": 360, "range": 664}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == {
            "DBZH",
            "VRADH",
            "KDP",
            "RHOHV",
            "PHIDP",
            "ZDR",
        }
        assert ds.sweep_number == 9

    # open last sweep group, auto
    with xr.open_dataset(
        iris0_file,
        group="sweep_9",
        engine="iris",
        backend_kwargs=dict(first_dim="time"),
    ) as ds:
        assert dict(ds.sizes) == {"time": 360, "range": 664}
        assert ds.sweep_number == 9


def test_open_iris1_dataset(iris1_file):
    # open first and only sweep group
    with xr.open_dataset(iris1_file, group="sweep_0", engine="iris") as ds:
        assert dict(ds.sizes) == {"azimuth": 359, "range": 833}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == {
            "DBZH",
            "KDP",
            "DBTH",
            "PHIDP",
            "ZDR",
            "RHOHV",
            "VRADH",
            "WRADH",
        }
        assert ds.sweep_number == 0

    # open first and only sweep group
    with xr.open_dataset(
        iris1_file,
        group="sweep_0",
        engine="iris",
        backend_kwargs=dict(first_dim="time"),
    ) as ds:
        assert dict(ds.sizes) == {"time": 359, "range": 833}
        assert ds.sweep_number == 0


@pytest.mark.parametrize(
    "compression, compression_opts", [("gzip", 0), ("gzip", 6), ("gzip", 9)]
)
def test_odim_roundtrip(odim_file2, compression, compression_opts):
    dtree = open_odim_datatree(odim_file2)
    with tempfile.NamedTemporaryFile(mode="w+b") as outfile:
        xradar.io.to_odim(
            dtree,
            outfile.name,
            source="WMO:01104,NOD:norst",
            compression=compression,
            compression_opts=compression_opts,
        )
        dtree2 = open_odim_datatree(outfile.name, reindex_angle=False)
        for d0, d1 in zip(dtree.groups, dtree2.groups):
            xr.testing.assert_equal(dtree[d0].ds, dtree2[d1].ds)


def test_odim_optional_how(odim_file2):
    dtree = open_odim_datatree(odim_file2)
    with tempfile.NamedTemporaryFile(mode="w+b") as outfile:
        xradar.io.to_odim(
            dtree,
            outfile.name,
            source="WMO:01104,NOD:norst",
            optional_how=True,
        )
        ds = h5py.File(outfile.name)

        for i in range(1, 6):
            ds_how = ds[f"dataset{i}"]["how"].attrs
            assert "scan_index" in ds_how
            assert "scan_count" in ds_how
            assert "startazA" in ds_how
            assert "stopazA" in ds_how
            assert "startazT" in ds_how
            assert "startazT" in ds_how
            assert "startelA" in ds_how
            assert "stopelA" in ds_how

    with tempfile.NamedTemporaryFile(mode="w+b") as outfile:
        xradar.io.to_odim(
            dtree,
            outfile.name,
            source="WMO:01104,NOD:norst",
            optional_how=False,
        )
        ds = h5py.File(outfile.name)

        for i in range(1, 6):
            ds_how = ds[f"dataset{i}"]["how"].attrs
            assert "scan_index" not in ds_how
            assert "scan_count" not in ds_how
            assert "startazA" not in ds_how
            assert "stopazA" not in ds_how
            assert "startazT" not in ds_how
            assert "startazT" not in ds_how
            assert "startelA" not in ds_how
            assert "stopelA" not in ds_how


def test_write_odim_source(rainbow_file2):
    dtree = open_rainbow_datatree(rainbow_file2)
    with tempfile.NamedTemporaryFile(mode="w+b") as outfile:
        with pytest.raises(ValueError):
            xradar.io.to_odim(
                dtree,
                outfile.name,
                source="PLC:Wideumont",
            )

        xradar.io.to_odim(
            dtree,
            outfile.name,
            source="NOD:bewid,WMO:06477",
        )
        ds = h5py.File(outfile.name)
        assert ds["what"].attrs["source"].decode("utf-8") == "NOD:bewid,WMO:06477"


def test_open_datamet_dataset(datamet_file):
    # open first sweep group
    with xr.open_dataset(
        datamet_file,
        group="sweep_0",
        engine="datamet",
    ) as ds:
        assert dict(ds.sizes) == {"azimuth": 360, "range": 493}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == {"DBTH", "DBZH", "KDP", "PHIDP", "RHOHV", "VRADH", "WRADH", "ZDR"}
        assert ds.sweep_number == 0

    # open last sweep group
    with xr.open_dataset(
        datamet_file,
        group="sweep_10",
        engine="datamet",
    ) as ds:
        assert dict(ds.sizes) == {"azimuth": 360, "range": 1332}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == {"DBTH", "DBZH", "KDP", "PHIDP", "RHOHV", "VRADH", "WRADH", "ZDR"}
        assert ds.sweep_number == 10


def test_open_datamet_dataset_reindex(datamet_file):
    # open first sweep group
    reindex_angle = dict(start_angle=0, stop_angle=360, angle_res=2.0, direction=1)
    with xr.open_dataset(
        datamet_file,
        group="sweep_10",
        engine="datamet",
        decode_coords=True,
        reindex_angle=reindex_angle,
    ) as ds:
        assert dict(ds.sizes) == {"azimuth": 180, "range": 1332}


def test_open_datamet_datatree(datamet_file):
    dtree = open_datamet_datatree(datamet_file)

    # root_attrs
    attrs = dtree.attrs
    assert attrs["Conventions"] == "None"

    # root vars
    rvars = dtree.data_vars
    assert rvars["volume_number"] == 0
    assert rvars["platform_type"] == "fixed"
    assert rvars["instrument_type"] == "radar"
    assert rvars["time_coverage_start"] == "2019-07-10T07:00:00Z"
    assert rvars["time_coverage_end"] == "2019-07-10T07:00:00Z"
    np.testing.assert_almost_equal(rvars["latitude"].values, np.array(41.9394))
    np.testing.assert_almost_equal(rvars["longitude"].values, np.array(14.6208))
    np.testing.assert_almost_equal(rvars["altitude"].values, np.array(710))

    # iterate over subgroups and check some values
    moments = ["DBTH", "DBZH", "KDP", "PHIDP", "RHOHV", "VRADH", "WRADH", "ZDR"]
    elevations = [16.1, 13.9, 11.0, 9.0, 7.0, 5.5, 4.5, 3.5, 2.5, 1.5, 0.5]
    azimuths = [360] * 11
    ranges = [493, 493, 493, 664, 832, 832, 1000, 1000, 1332, 1332, 1332]
    i = 0
    for grp in dtree.match("sweep_*"):
        ds = dtree[grp].ds
        assert dict(ds.sizes) == {"azimuth": azimuths[i], "range": ranges[i]}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == set(moments)
        assert set(ds.data_vars) & (required_sweep_metadata_vars) == set(
            required_sweep_metadata_vars ^ {"azimuth", "elevation"}
        )
        assert set(ds.coords) == {
            "azimuth",
            "elevation",
            "time",
            "latitude",
            "longitude",
            "altitude",
            "range",
        }
        assert np.round(ds.elevation.mean().values.item(), 1) == elevations[i]
        assert ds.sweep_number == i
        i += 1

    # Try to reed single sweep
    dtree = open_datamet_datatree(datamet_file, sweep=1)
    assert len(dtree.groups) == 5

    # Try to read list of sweeps
    dtree = open_datamet_datatree(datamet_file, sweep=[1, 2])
    assert len(dtree.groups) == 6


@pytest.mark.parametrize("first_dim", ["time", "auto"])
def test_cfradfial2_roundtrip(cfradial1_file, first_dim):
    dtree0 = open_cfradial1_datatree(cfradial1_file, first_dim=first_dim)
    # first write to cfradial2
    with tempfile.NamedTemporaryFile(mode="w+b") as outfile:
        xradar.io.to_cfradial2(dtree0.copy(), outfile.name)
        # then open cfradial2 file
        dtree1 = xr.open_datatree(outfile.name)
        # and write again
        with tempfile.NamedTemporaryFile(mode="w+b") as outfile1:
            xradar.io.to_cfradial2(dtree1.copy(), outfile1.name)
            # and open second cfradial2
            dtree2 = xr.open_datatree(outfile1.name)
            # check equality
            for d0, d1, d2 in zip(dtree0.groups, dtree1.groups, dtree2.groups):
                if "sweep" in d0:
                    if first_dim == "auto":
                        first_dim = "azimuth"
                    assert first_dim in dtree0[d0].dims
                    assert "time" in dtree1[d1].dims
                    assert "time" in dtree2[d2].dims
                xr.testing.assert_equal(dtree1[d1].ds, dtree2[d2].ds)


def test_cfradial_n_points_file(cfradial1n_file):
    dtree = open_cfradial1_datatree(
        cfradial1n_file, first_dim="auto", site_coords=False
    )
    attrs = dtree.attrs

    # root_attrs
    assert attrs["Conventions"] == "CF-1.7"
    assert attrs["version"] == "CF-Radial-1.4"
    assert attrs["title"] == "VOL_A"
    assert attrs["instrument_name"] == "Desio_Radar"
    assert attrs["platform_is_mobile"] == "false"

    # root vars
    rvars = dtree.data_vars
    assert rvars["volume_number"] == 1
    assert rvars["platform_type"] == b"fixed"
    assert rvars["instrument_type"] == b"radar"
    assert rvars["time_coverage_start"] == b"2024-05-22T16:00:47Z"
    assert rvars["time_coverage_end"] == b"2024-05-22T16:03:20Z"
    np.testing.assert_almost_equal(rvars["latitude"].values, np.array(45.6272661))
    np.testing.assert_almost_equal(rvars["longitude"].values, np.array(9.1963181))
    np.testing.assert_almost_equal(rvars["altitude"].values, np.array(241.0))

    # iterate over subgroups and check some values
    moments = ["ZDR", "RHOHV", "KDP", "DBZ", "VEL", "PHIDP"]
    elevations = [0.7, 1.3, 3.0, 5.0, 7.0, 10.0, 15.0, 25.0]
    azimuths = [360] * 8
    ranges = [416] * 5 + [383, 257, 157]
    for grp in dtree.groups:
        # only iterate sweep groups
        if "sweep" not in grp:
            continue
        ds = dtree[grp].ds
        i = int(ds.sweep_number.values)
        assert i == int(grp[7:])
        assert dict(ds.sizes) == {
            "azimuth": azimuths[i],
            "range": ranges[i],
            "frequency": 1,
        }
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == set(moments)
        assert set(ds.data_vars) & (required_sweep_metadata_vars) == set(
            required_sweep_metadata_vars ^ {"azimuth", "elevation"}
        )
        assert set(ds.coords) == {
            "azimuth",
            "elevation",
            "time",
            "range",
            "frequency",
        }
        assert np.round(ds.sweep_fixed_angle.values.item(), 1) == elevations[i]
        assert ds.sweep_mode == "azimuth_surveillance"


@pytest.mark.run(order=1)
@pytest.mark.parametrize("sweep", ["sweep_0", 0, [0, 1], ["sweep_0", "sweep_1"]])
@pytest.mark.parametrize(
    "nexradlevel2_files", ["nexradlevel2_gzfile", "nexradlevel2_bzfile"], indirect=True
)
def test_open_nexradlevel2_datatree_sweep(nexradlevel2_files, sweep):
    dtree = open_nexradlevel2_datatree(nexradlevel2_files, sweep=sweep)
    if isinstance(sweep, (str, int)):
        lswp = len([sweep])
    else:
        lswp = len(sweep)
    assert len(dtree.match("sweep*")) == lswp


@pytest.mark.parametrize(
    "nexradlevel2_files", ["nexradlevel2_gzfile", "nexradlevel2_bzfile"], indirect=True
)
def test_open_nexradlevel2_datatree(nexradlevel2_files):
    dtree = open_nexradlevel2_datatree(nexradlevel2_files)
    # root_attrs
    attrs = dtree.attrs
    assert attrs["Conventions"] == "None"
    assert attrs["instrument_name"] == "KLBB"

    # root vars
    rvars = dtree.data_vars
    assert rvars["volume_number"] == 0
    assert rvars["platform_type"] == "fixed"
    assert rvars["instrument_type"] == "radar"
    assert rvars["time_coverage_start"] == "2016-06-01T15:00:25Z"
    assert rvars["time_coverage_end"] == "2016-06-01T15:06:06Z"
    np.testing.assert_almost_equal(rvars["latitude"].values, np.array(33.65414047))
    np.testing.assert_almost_equal(rvars["longitude"].values, np.array(-101.81416321))
    np.testing.assert_almost_equal(rvars["altitude"].values, np.array(1029))

    # iterate over subgroups and check some values
    moments = [
        ["DBZH", "PHIDP", "RHOHV", "ZDR"],
        ["DBZH", "WRADH", "VRADH"],
        ["DBZH", "PHIDP", "RHOHV", "ZDR"],
        ["DBZH", "WRADH", "VRADH"],
        ["DBZH", "PHIDP", "RHOHV", "VRADH", "WRADH", "ZDR"],
        ["DBZH", "PHIDP", "RHOHV", "VRADH", "WRADH", "ZDR"],
        ["DBZH", "PHIDP", "RHOHV", "VRADH", "WRADH", "ZDR"],
        ["DBZH", "PHIDP", "RHOHV", "VRADH", "WRADH", "ZDR"],
        ["DBZH", "PHIDP", "RHOHV", "VRADH", "WRADH", "ZDR"],
        ["DBZH", "PHIDP", "RHOHV", "VRADH", "WRADH", "ZDR"],
        ["DBZH", "PHIDP", "RHOHV", "VRADH", "WRADH", "ZDR"],
    ]
    elevations = [
        0.5,
        0.5,
        1.5,
        1.5,
        2.4,
        3.4,
        4.3,
        6.0,
        9.9,
        14.6,
        19.5,
    ]
    azimuths = [
        720,
        720,
        720,
        720,
        360,
        360,
        360,
        360,
        360,
        360,
        360,
    ]
    ranges = [
        1832,
        1192,
        1632,
        1192,
        1312,
        1076,
        908,
        696,
        448,
        308,
        232,
    ]
    assert len(dtree.groups[1:]) == 14
    for i, grp in enumerate(dtree.match("sweep_*")):
        print(i)
        ds = dtree[grp].ds
        assert dict(ds.sizes) == {"azimuth": azimuths[i], "range": ranges[i]}
        assert set(ds.data_vars) & (
            sweep_dataset_vars | non_standard_sweep_dataset_vars
        ) == set(moments[i])
        assert set(ds.data_vars) & (required_sweep_metadata_vars) == set(
            required_sweep_metadata_vars ^ {"azimuth", "elevation"}
        )
        assert set(ds.coords) == {
            "azimuth",
            "elevation",
            "time",
            "latitude",
            "longitude",
            "altitude",
            "range",
        }
        assert np.round(ds.elevation.mean().values.item(), 1) == elevations[i]
        assert ds.sweep_number.values == int(grp[6:])


@skip_import("dask")
@pytest.mark.parametrize(
    "nexradlevel2_files", ["nexradlevel2_gzfile", "nexradlevel2_bzfile"], indirect=True
)
def test_nexradlevel2_dask_load(nexradlevel2_files):
    ds = xr.open_dataset(nexradlevel2_files, group="sweep_0", engine="nexradlevel2")
    dsc = ds.chunk()
    dsc.load()


@skip_import("dask")
def test_iris_dask_load(iris0_file):
    ds = xr.open_dataset(iris0_file, group="sweep_0", engine="iris")
    dsc = ds.chunk()
    dsc.load()



#!/usr/bin/env python
# Copyright (c) 2024, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.
"""
Transform CF-Radial
===================

This module provides the utilities to convert between CfRadial1 and
CfRadial2 formats, following WMO guidelines for radar data processing.

.. autosummary::
   :nosignatures:
   :toctree: generated/

    {}

to_cfradial1
------------
Convert a `xarray.DataTree` object to CfRadial1 format. The output is
an `xarray.Dataset` following the CfRadial1 standard, which can be used
for radar data visualization and further processing.

Parameters:
    - dtree: xarray.DataTree
        Radar xarray object to convert.
    - filename: str (optional)
        Output file name for the CfRadial1 dataset.
    - calibs: bool (default: True)
        Whether to include calibration parameters.

Returns:
    - dataset: xarray.Dataset
        Converted dataset in CfRadial1 format.

to_cfradial2
------------
Convert a `xarray.Dataset` object to CfRadial2 format by mapping the
structure back into a `xarray.DataTree`.

Parameters:
    - ds: xarray.Dataset
        CfRadial1 dataset to convert.
    - kwargs: dict
        Additional keyword arguments for controlling the conversion, such
        as specifying sweeps or metadata inclusion.

Returns:
    - dtree: xarray.DataTree
        Converted data tree in CfRadial2 format.

"""

__all__ = [
    "to_cfradial2",
    "to_cfradial1",
]

__doc__ = __doc__.format("\n   ".join(__all__))

from importlib.metadata import version

from xarray import DataTree

from ..io.backends.cfradial1 import (
    _get_radar_calibration,
    _get_required_root_dataset,
    _get_subgroup,
    _get_sweep_groups,
)
from ..io.backends.common import _attach_sweep_groups
from ..io.export.cfradial1 import (
    _calib_mapper,
    _variable_mapper,
)
from ..model import (
    georeferencing_correction_subgroup,
    radar_parameters_subgroup,
)


# to_cfradial1 function implementation
def to_cfradial1(dtree=None, calibs=True):
    """
    Convert a radar xarray.DataTree to the CFRadial1 format
    and save it to a file. Ensure that the resulting dataset
    is well-formed and does not include specified extraneous variables.

    Parameters
    ----------
    dtree: xarray.DataTree
        Radar xarray.DataTree object.
    calibs: Bool, optional
        Whether to include calibration parameters.
    """
    # Generate the initial ds_cf using the existing mapping functions
    dataset = _variable_mapper(dtree)

    # Handle calibration parameters
    if calibs:
        if "radar_calibration" in dtree:
            calib_params = dtree["radar_calibration"].to_dataset()
            calibs = _calib_mapper(calib_params)
            dataset.update(calibs)

    # Add additional parameters if they exist in dtree
    if "radar_parameters" in dtree:
        radar_params = dtree["radar_parameters"].to_dataset().reset_coords()
        dataset.update(radar_params)

    if "georeferencing_correction" in dtree:
        radar_georef = dtree["georeferencing_correction"].to_dataset().reset_coords()
        dataset.update(radar_georef)

    # Ensure that the data type of sweep_mode and similar variables matches
    if "sweep_mode" in dataset.variables:
        dataset["sweep_mode"] = dataset["sweep_mode"].astype("S")

    # Update global attributes
    dataset.attrs = dtree.attrs
    dataset.attrs["Conventions"] = "Cf/Radial"
    dataset.attrs["version"] = "1.2"
    xradar_version = version("xradar")
    dataset.attrs["history"] += f": xradar v{xradar_version} CfRadial1 export"

    return dataset


# to_cfradial2 function implementation
def to_cfradial2(ds, **kwargs):
    """Convert a CfRadial1 Dataset to CfRadial2 DataTree."""
    first_dim = kwargs.pop("first_dim", "auto")
    optional = kwargs.pop("optional", True)
    kwargs.pop("site_coords", True)
    sweep = kwargs.pop("sweep", None)

    # Create DataTree root node with required data
    root_data = _get_required_root_dataset(ds, optional=optional)
    dtree = DataTree(dataset=root_data, name="root")

    # Attach additional root metadata groups as child nodes
    radar_parameters = _get_subgroup(ds, radar_parameters_subgroup)
    if radar_parameters:
        dtree["radar_parameters"] = DataTree(radar_parameters, name="radar_parameters")

    calib = _get_radar_calibration(ds)
    if calib:
        dtree["radar_calibration"] = DataTree(calib, name="radar_calibration")

    georeferencing = _get_subgroup(ds, georeferencing_correction_subgroup)
    if georeferencing:
        dtree["georeferencing_correction"] = DataTree(
            georeferencing, name="georeferencing_correction"
        )

    # Attach sweep child nodes
    sweep_groups = list(
        _get_sweep_groups(
            ds,
            sweep=sweep,
            first_dim=first_dim,
            optional=optional,
            site_coords=True,
        ).values()
    )
    dtree = _attach_sweep_groups(dtree, sweep_groups)

    return dtree



#!/usr/bin/env python
# Copyright (c) 2024, openradar developers.
# Distributed under the MIT License. See LICENSE for more info.

"""
Metek MRR2 raw and processed data
=================================
Read data from METEK's MRR-2 raw (.raw) and processed (.pro, .avg) files.

Example::

   import xradar as xd
   ds = xr.open_dataset('0308.pro', engine='metek')

.. autosummary::
   :nosignatures:
   :toctree: generated/

   {}
"""

import io
import warnings
from datetime import datetime

import numpy as np
import xarray as xr
from xarray import DataTree
from xarray.backends.common import AbstractDataStore, BackendArray, BackendEntrypoint
from xarray.backends.file_manager import CachingFileManager
from xarray.backends.store import StoreBackendEntrypoint
from xarray.core import indexing
from xarray.core.utils import FrozenDict

from ...model import (
    georeferencing_correction_subgroup,
    get_altitude_attrs,
    get_azimuth_attrs,
    get_elevation_attrs,
    get_latitude_attrs,
    get_longitude_attrs,
    get_time_attrs,
    radar_calibration_subgroup,
    radar_parameters_subgroup,
)
from .common import (
    _attach_sweep_groups,
    _get_radar_calibration,
    _get_required_root_dataset,
    _get_subgroup,
)

__all__ = [
    "MRRBackendEntrypoint",
    "open_metek_datatree",
]

__doc__ = __doc__.format("\n   ".join(__all__))

variable_attr_dict = dict(
    transfer_function={
        "long_name": "Transfer function",
        "standard_name": "transfer_function",
        "units": "1",
        "dims": ("time", "range"),
    },
    spectral_reflectivity={
        "long_name": "Spectral reflectivity",
        "standard_name": "equivalent_reflectivity_factor",
        "units": "dB",
        "dims": ("index", "sample"),
    },
    raw_spectra_counts={
        "long_name": "Raw spectra counts",
        "standard_name": "raw_spectra",
        "units": "",
        "dims": ("index", "sample"),
    },
    drop_size={
        "long_name": "Drop size",
        "standard_name": "drop_size",
        "units": "mm",
        "dims": ("index", "sample"),
    },
    drop_number_density={
        "long_name": "Raindrop number density",
        "standard_name": "raindrop_number_density",
        "units": "m-4",
        "dims": ("index", "sample"),
    },
    rainfall_rate={
        "long_name": "Rainfall rate",
        "standard_name": "rainfall_rate",
        "units": "mm hr-1",
        "dims": ("time", "range"),
    },
    liquid_water_content={
        "long_name": "Liquid water content",
        "standard_name": "liquid_water_content",
        "units": "g m-3",
        "dims": ("time", "range"),
    },
    path_integrated_attenuation={
        "long_name": "Path integrated attenuation",
        "standard_name": "path_integrated_attenuation",
        "units": "dB",
        "dims": ("time", "range"),
    },
    corrected_reflectivity={
        "long_name": "Attenuation-corrected Radar reflectivity factor",
        "standard_name": "equivalent_radar_reflectivity_factor",
        "units": "dBZ",
        "dims": ("time", "range"),
    },
    reflectivity={
        "long_name": "Radar reflectivity factor",
        "standard_name": "equivalent_radar_reflectivity_factor",
        "units": "dBZ",
        "dims": ("time", "range"),
    },
    spectrum_index={
        "long_name": "Spectrum index",
        "standard_name": "spectrum_index",
        "units": "1",
        "dims": ("time", "range"),
    },
    percentage_valid_spectra={
        "long_name": "Percentage of spectra that are valid",
        "standard_name": "percentage_valid_spectra",
        "units": "percent",
        "dims": ("time"),
    },
    number_valid_spectra={
        "long_name": "number of spectra that are valid",
        "standard_name": "number_valid_spectra",
        "units": "1",
        "dims": ("time"),
    },
    total_number_spectra={
        "long_name": "Total number of spectra",
        "standard_name": "total_number_spectra",
        "units": "1",
        "dims": ("time"),
    },
    velocity_bins={
        "long_name": "Doppler velocity bins",
        "standard_name": "doppler_velocity_bins",
        "units": "m s-1",
        "dims": ("sample"),
    },
    range={
        "long_name": "Range from radar",
        "standard_name": "range",
        "units": "m",
        "dims": ("range",),
    },
    time=get_time_attrs(),
    azimuth=get_azimuth_attrs(),
    elevation=get_elevation_attrs(),
    velocity={
        "long_name": "Radial velocity of scatterers toward instrument",
        "standard_name": "radial_velocity_of_scatterers_toward_instrument",
        "units": "m s-1",
    },
    latitude=get_latitude_attrs(),
    longitude=get_longitude_attrs(),
    altitude=get_altitude_attrs(),
)

variable_attr_dict["time"]["dims"] = ("time",)
variable_attr_dict["azimuth"]["dims"] = ("time",)
variable_attr_dict["elevation"]["dims"] = ("time",)
variable_attr_dict["velocity"]["dims"] = ("time", "range")
variable_attr_dict["latitude"]["dims"] = ()
variable_attr_dict["longitude"]["dims"] = ()
variable_attr_dict["altitude"]["dims"] = ()


def _parse_spectra_line(input_str, num_gates):
    out_array = np.zeros(num_gates)
    increment = {32: 9, 31: 7}[num_gates]
    for i, pos in enumerate(range(3, len(input_str) - increment, increment)):
        input_num_str = input_str[pos : pos + increment]
        try:
            out_array[i] = float(input_num_str)
        except ValueError:
            out_array[i] = np.nan

    return out_array


class MRR2File:
    def __init__(self, file_name="", **kwargs):
        self.vel_bin_spacing = 0.1887
        self.nyquist_velocity = self.vel_bin_spacing * 64
        self._data = {}
        self._data["velocity_bins"] = np.arange(
            0, 64 * self.vel_bin_spacing, self.vel_bin_spacing
        )
        self._data["range"] = []
        self._data["transfer_function"] = []
        self._data["spectral_reflectivity"] = []
        self._data["raw_spectra_counts"] = []
        self._data["drop_size"] = []
        self._data["drop_number_density"] = []
        self._data["time"] = []
        self.filetype = ""
        self.device_version = ""
        self.device_serial_number = ""
        self.bandwidth = 0
        self._fp = None
        self.calibration_constant = []
        self._data["percentage_valid_spectra"] = []
        self._data["number_valid_spectra"] = []
        self._data["total_number_spectra"] = []
        self.spectra_index = 0
        self.altitude = None
        self.sampling_rate = 125000.0
        self._data["path_integrated_attenuation"] = []
        self._data["corrected_reflectivity"] = []
        self._data["reflectivity"] = []
        self._data["rainfall_rate"] = []
        self._data["liquid_water_content"] = []
        self._data["velocity"] = []
        self._data["altitude"] = np.array(np.nan)
        self._data["longitude"] = np.array(np.nan)
        self._data["latitude"] = np.array(np.nan)
        self.filename = None
        self.n_gates = 32
        if not file_name == "":
            self.filename = file_name
            self.open(file_name)

    def open(self, filename_or_obj):
        if isinstance(filename_or_obj, io.IOBase):
            filename_or_obj.seek(0)
            self._fp = filename_or_obj

        if isinstance(filename_or_obj, str):
            self.filename = filename_or_obj
            self._fp = open(filename_or_obj)

        num_times = 0
        temp_spectra = np.zeros((self.n_gates, 64))
        temp_drops = np.zeros((self.n_gates, 64))
        temp_number = np.zeros((self.n_gates, 64))
        spec_var = ""
        for file_line in self._fp:
            if isinstance(file_line, bytes):
                file_line = file_line.decode("utf-8")
            if file_line[:3] == "MRR":
                if num_times > 0:
                    self._data[spec_var].append(temp_spectra)
                    self._data["drop_number_density"].append(temp_number)
                    self._data["drop_size"].append(temp_drops)

                string_split = file_line.split()
                time_str = string_split[1]
                parsed_datetime = datetime.strptime(time_str, "%y%m%d%H%M%S")
                self._data["time"] = self._data["time"] + [parsed_datetime]
                self.filetype = string_split[-1]
                if self.filetype == "RAW":
                    self.device_version = string_split[4]
                    self.device_serial_number = string_split[6]
                    self.bandwidth = int(string_split[8])
                    self.calibration_constant.append(int(string_split[10]))
                    self._data["percentage_valid_spectra"].append(int(string_split[12]))
                    self._data["number_valid_spectra"].append(int(string_split[13]))
                    self._data["total_number_spectra"].append(int(string_split[14]))
                    self.n_gates = 32
                    spec_var = "raw_spectra_counts"
                elif self.filetype == "AVE" or self.filetype == "PRO":
                    self._data["altitude"] = np.array(float(string_split[8]))
                    self.sampling_rate = float(string_split[10])
                    self.mrr_service_version = string_split[12]
                    self.device_version = string_split[14]
                    self.calibration_constant.append(int(string_split[16]))
                    self._data["percentage_valid_spectra"].append(int(string_split[18]))
                    self.n_gates = 31
                    spec_var = "spectral_reflectivity"
                else:
                    raise OSError(
                        "Invalid file type flag in file! Must be RAW, AVG, or PRO!"
                    )
                temp_spectra = np.zeros((self.n_gates, 64))
                temp_drops = np.zeros((self.n_gates, 64))
                temp_number = np.zeros((self.n_gates, 64))
                num_times = num_times + 1

            if file_line[0] == "H":
                in_array = _parse_spectra_line(file_line, self.n_gates)
                if num_times > 1:
                    after_res = in_array[1] - in_array[0]
                    before_res = self._data["range"][1] - self._data["range"][0]
                    if not after_res == before_res:
                        warnings.warn(
                            f"MRR2 resolution was changed mid file. Before time period "
                            f"{parsed_datetime} the resolution was {before_res}, "
                            f"and {after_res} after.",
                            UserWarning,
                        )
                self._data["range"] = in_array

            if file_line[0:2] == "TF":
                self._data["transfer_function"].append(
                    _parse_spectra_line(file_line, self.n_gates)
                )
            if file_line[0] == "F":
                spectra_bin_no = int(file_line[1:3])
                temp_spectra[:, spectra_bin_no] = _parse_spectra_line(
                    file_line, self.n_gates
                )
            if file_line[0] == "D":
                spectra_bin_no = int(file_line[1:3])
                temp_drops[:, spectra_bin_no] = _parse_spectra_line(
                    file_line, self.n_gates
                )
            if file_line[0] == "N":
                spectra_bin_no = int(file_line[1:3])
                temp_number[:, spectra_bin_no] = _parse_spectra_line(
                    file_line, self.n_gates
                )
            if file_line[0:3] == "PIA":
                self._data["path_integrated_attenuation"] = self._data[
                    "path_integrated_attenuation"
                ] + [_parse_spectra_line(file_line, self.n_gates)]
            if file_line[0:3] == "z  ":
                self._data["reflectivity"] = self._data["reflectivity"] + [
                    _parse_spectra_line(file_line, self.n_gates)
                ]
            if file_line[0:3] == "Z  ":
                self._data["corrected_reflectivity"] = self._data[
                    "corrected_reflectivity"
                ] + [_parse_spectra_line(file_line, self.n_gates)]
            if file_line[0:3] == "RR ":
                self._data["rainfall_rate"] = self._data["rainfall_rate"] + [
                    _parse_spectra_line(file_line, self.n_gates)
                ]
            if file_line[0:3] == "LWC":
                self._data["liquid_water_content"] = self._data[
                    "liquid_water_content"
                ] + [_parse_spectra_line(file_line, self.n_gates)]
            if file_line[0:3] == "W  ":
                self._data["velocity"] = self._data["velocity"] + [
                    _parse_spectra_line(file_line, self.n_gates)
                ]

        self._data[spec_var].append(temp_spectra)
        self._data["drop_number_density"].append(temp_number)
        self._data["drop_size"].append(temp_drops)
        self._data["transfer_function"] = np.stack(
            self._data["transfer_function"], axis=0
        )
        self._data[spec_var] = np.stack(self._data[spec_var], axis=0)
        self._data["drop_number_density"] = np.stack(
            self._data["drop_number_density"], axis=0
        )
        self._data["drop_size"] = np.stack(self._data["drop_size"], axis=0)

        if self.filetype == "RAW":
            self._data["total_number_spectra"] = np.stack(
                self._data["total_number_spectra"], axis=0
            )
            self._data["number_valid_spectra"] = np.stack(
                self._data["number_valid_spectra"], axis=0
            )
            self._data["reflectivity"] = None
            self._data["corrected_reflectivity"] = None
            self._data["liquid_water_content"] = None
            self._data["rainfall_rate"] = None
            self._data["percentage_valid_spectra"] = None
            self._data["drop_number_density"] = None
            self._data["drop_size"] = None
            self._data["path_integrated_attenuation"] = None
            self._data["velocity"] = None
            self._data["spectral_reflectivity"] = None

            del self._data["reflectivity"]
            del self._data["corrected_reflectivity"]
            del self._data["liquid_water_content"]
            del self._data["rainfall_rate"]
            del self._data["percentage_valid_spectra"]
            del self._data["drop_number_density"]
            del self._data["drop_size"]
            del self._data["path_integrated_attenuation"]
            del self._data["velocity"]
            del self._data["spectral_reflectivity"]
        else:
            del self._data["total_number_spectra"], self._data["number_valid_spectra"]
            self._data["reflectivity"] = np.stack(self._data["reflectivity"], axis=0)
            self._data["path_integrated_attenuation"] = np.stack(
                self._data["path_integrated_attenuation"], axis=0
            )
            self._data["corrected_reflectivity"] = np.stack(
                self._data["corrected_reflectivity"], axis=0
            )

            self._data["liquid_water_content"] = np.stack(
                self._data["liquid_water_content"], axis=0
            )
            self._data["velocity"] = np.stack(self._data["velocity"], axis=0)
            self._data["rainfall_rate"] = np.stack(self._data["rainfall_rate"], axis=0)
            self._data["percentage_valid_spectra"] = np.stack(
                self._data["percentage_valid_spectra"], axis=0
            )
            self._data["drop_number_density"] = np.stack(
                self._data["drop_number_density"], axis=0
            )
            self._data["drop_size"] = np.stack(self._data["drop_size"], axis=0)
            self._data["raw_spectra_counts"] = None
            del self._data["raw_spectra_counts"]

        self._data["range"] = np.squeeze(self._data["range"])
        # Now we compress the spectrum variables to remove invalid spectra
        self._data[spec_var] = self._data[spec_var].reshape(
            self._data[spec_var].shape[0] * self._data[spec_var].shape[1],
            self._data[spec_var].shape[2],
        )
        where_valid_spectra = np.any(np.isfinite(self._data[spec_var]), axis=1)
        inds = np.where(where_valid_spectra, 1, -1)

        self._data[spec_var] = self._data[spec_var][where_valid_spectra]
        if self.filetype == "PRO" or self.filetype == "AVE":
            self._data["drop_number_density"] = self._data[
                "drop_number_density"
            ].reshape(
                self._data["drop_number_density"].shape[0]
                * self._data["drop_number_density"].shape[1],
                self._data["drop_number_density"].shape[2],
            )
            self._data["drop_number_density"] = self._data["drop_number_density"][
                where_valid_spectra
            ]
            self._data["drop_size"] = self._data["drop_size"].reshape(
                self._data["drop_size"].shape[0] * self._data["drop_size"].shape[1],
                self._data["drop_size"].shape[2],
            )
            self._data["drop_size"] = self._data["drop_size"][where_valid_spectra]
        cur_index = 0
        for i in range(len(inds)):
            if inds[i] > -1:
                inds[i] = cur_index
                cur_index += 1
        self._data["spectrum_index"] = inds.reshape(
            (len(self._data["time"]), len(self._data["range"]))
        )

        self._data["azimuth"] = np.zeros_like(self._data["time"])
        self._data["elevation"] = 90 * np.ones_like(self._data["time"])
        self._data["time"] = np.array(self._data["time"])
        temp_drops = None
        temp_number = None
        temp_spectra = None

    def close(self):
        if self._fp is not None:
            self._fp.close()

    __del__ = close

    @property
    def data(self):
        return self._data

    @property
    def coordinates(self):
        return self._coordinates

    @property
    def fixed_angle(self):
        return self._data["elevation"]

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


class MRR2ArrayWrapper(BackendArray):
    def __init__(
        self,
        data,
    ):
        self.data = data
        self.shape = data.shape
        self.dtype = np.dtype("float64")

    def __getitem__(self, key: tuple):
        return indexing.explicit_indexing_adapter(
            key,
            self.shape,
            indexing.IndexingSupport.OUTER_1VECTOR,
            self._raw_indexing_method,
        )

    def _raw_indexing_method(self, key: tuple):
        return self.data[key]


class MRR2DataStore(AbstractDataStore):
    def __init__(self, manager, group=None):
        self._manager = manager
        self._group = group
        self._filename = self.filename
        self._need_time_recalc = False

    @classmethod
    def open(cls, filename, mode="r", group=None, **kwargs):
        manager = CachingFileManager(MRR2File, filename, mode=mode, kwargs=kwargs)
        return cls(manager, group=group)

    @property
    def filename(self):
        with self._manager.acquire_context(False) as root:
            return root.filename

    @property
    def root(self):
        with self._manager.acquire_context(False) as root:
            return root

    def _acquire(self, needs_lock=True):
        with self._manager.acquire_context(needs_lock) as root:
            return root

    @property
    def ds(self):
        return self._acquire()

    def open_store_variable(self, name, var):
        data = indexing.LazilyOuterIndexedArray(MRR2ArrayWrapper(var))
        encoding = {"group": self._group, "source": self._filename}
        attrs = variable_attr_dict[name].copy()
        dims = attrs["dims"]
        del attrs["dims"]
        return xr.Variable(dims, data, attrs, encoding)

    def open_store_coordinates(self):
        coord_keys = ["time", "range", "velocity_bins"]
        coords = {}
        for k in coord_keys:
            attrs = variable_attr_dict[k].copy()
            dims = attrs["dims"]
            del attrs["dims"]
            coords[k] = xr.Variable(dims, self.ds.data[k], attrs=attrs)

        return coords

    def get_variables(self):
        return FrozenDict(
            (k1, v1)
            for k1, v1 in {
                **{k: self.open_store_variable(k, v) for k, v in self.ds.data.items()},
                **self.open_store_coordinates(),
            }.items()
        )

    def get_attrs(self):
        return FrozenDict()


class MRRBackendEntrypoint(BackendEntrypoint):
    """Xarray BackendEntrypoint for Metek MRR2 data.

    Keyword Arguments
    -----------------
    first_dim : str
        Can be ``time`` or ``auto`` first dimension. If set to ``auto``,
        first dimension will be either ``azimuth`` or ``elevation`` depending on
        type of sweep. Defaults to ``auto``.
    site_coords : bool
        Attach radar site-coordinates to Dataset, defaults to ``True``.
    kwargs : dict
        Additional kwargs are fed to :py:func:`xarray.open_dataset`.
    """

    description = "Backend for reading Metek MRR2 processed and raw data"
    url = "https://xradar.rtfd.io/en/latest/io.html#metek"

    def open_dataset(
        self,
        filename_or_obj,
        *,
        mask_and_scale=True,
        decode_times=True,
        concat_characters=True,
        decode_coords=True,
        drop_variables=None,
        use_cftime=None,
        decode_timedelta=None,
        format=None,
        group="/",
        invalid_netcdf=None,
        phony_dims="access",
        decode_vlen_strings=True,
        first_dim="auto",
        site_coords=True,
        optional=True,
    ):
        store_entrypoint = StoreBackendEntrypoint()

        store = MRR2DataStore.open(
            filename_or_obj,
            format=format,
            group=group,
            invalid_netcdf=invalid_netcdf,
            phony_dims=phony_dims,
            decode_vlen_strings=decode_vlen_strings,
        )

        ds = store_entrypoint.open_dataset(
            store,
            mask_and_scale=mask_and_scale,
            decode_times=decode_times,
            concat_characters=concat_characters,
            decode_coords=decode_coords,
            drop_variables=drop_variables,
            use_cftime=use_cftime,
            decode_timedelta=decode_timedelta,
        )

        ds = ds.assign_coords({"range": ds.range})
        ds = ds.assign_coords({"time": ds.time})
        ds = ds.assign_coords({"velocity_bins": ds.velocity_bins})
        ds.encoding["engine"] = "metek"

        return ds


def open_metek_datatree(filename_or_obj, **kwargs):
    """Open Metek MRR2 dataset as :py:class:`xarray.DataTree`.

    Parameters
    ----------
    filename_or_obj : str, Path, file-like or DataStore
        Strings and Path objects are interpreted as a path to a local or remote
        radar file

    Keyword Arguments
    -----------------
    sweep : int, list of int, optional
        Sweep number(s) to extract, default to first sweep. If None, all sweeps are
        extracted into a list.
    first_dim : str
        Can be ``time`` or ``auto`` first dimension. If set to ``auto``,
        first dimension will be either ``azimuth`` or ``elevation`` depending on
        type of sweep. Defaults to ``auto``.
    reindex_angle : bool or dict
        Defaults to False, no reindexing. Given dict should contain the kwargs to
        reindex_angle. Only invoked if `decode_coord=True`.
    fix_second_angle : bool
        If True, fixes erroneous second angle data. Defaults to ``False``.
    site_coords : bool
        Attach radar site-coordinates to Dataset, defaults to ``True``.
    kwargs : dict
        Additional kwargs are fed to :py:func:`xarray.open_dataset`.

    Returns
    -------
    dtree: xarray.DataTree
        DataTree
    """
    # handle kwargs, extract first_dim
    backend_kwargs = kwargs.pop("backend_kwargs", {})
    optional = backend_kwargs.pop("optional", True)
    sweep = kwargs.pop("sweep", None)
    sweeps = []
    kwargs["backend_kwargs"] = backend_kwargs

    if isinstance(sweep, str):
        sweeps = [sweep]
    elif isinstance(sweep, int):
        sweeps = [f"sweep_{sweep}"]
    elif isinstance(sweep, list):
        if isinstance(sweep[0], int):
            sweeps = [f"sweep_{i + 1}" for i in sweep]
        else:
            sweeps.extend(sweep)
    else:
        sweeps = ["sweep_0"]

    ls_ds: list[xr.Dataset] = [
        xr.open_dataset(filename_or_obj, group=swp, engine="metek", **kwargs)
        for swp in sweeps
    ].copy()
    dtree: dict = {
        "/": _get_required_root_dataset(ls_ds, optional=optional),
        "/radar_parameters": _get_subgroup(ls_ds, radar_parameters_subgroup),
        "/georeferencing_correction": _get_subgroup(
            ls_ds, georeferencing_correction_subgroup
        ),
        "/radar_calibration": _get_radar_calibration(ls_ds, radar_calibration_subgroup),
    }
    dtree = _attach_sweep_groups(dtree, ls_ds)
    return DataTree.from_dict(dtree)
